# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# This code was inspired in the autotest project,
#
# client/base_utils.py
# Original author: Martin J Bligh <mbligh@google.com>
# Original author: John Admanski <jadmanski@google.com>


"""
Get information from the current's machine CPU.
"""

import glob
import logging
import os
import platform
import random
import re
import warnings

#: Map vendor's name with expected string in /proc/cpuinfo.
VENDORS_MAP = {
    'intel': (b"GenuineIntel", ),
    'amd': (b"AMD", ),
    'ibm': (rb"POWER\d", rb"IBM/S390", ),
}


class FamilyException(Exception):
    pass


def _list_matches(content_list, pattern):
    """
    Checks if any item in list matches the specified pattern.

    :param content_list: items to match
    :type content_list: list
    :param pattern: pattern to match from content_list
    :type pattern: str
    :return: if the pattern was found or not
    :rtype: bool
    """
    for content in content_list:
        match = re.search(pattern, content)
        if match:
            return True
    return False


def _get_info():
    """
    Returns info on the 1st CPU entry from /proc/cpuinfo as a list of lines.

    :return: `list` of lines 1st CPU entry from /proc/cpuinfo file
    :rtype: list
    """
    cpuinfo = []
    with open('/proc/cpuinfo', 'rb') as proc_cpuinfo:
        for line in proc_cpuinfo:
            if line == b'\n':
                break
            cpuinfo.append(line)
    return cpuinfo


def _get_status(cpu):
    """
    Check if a CPU is online or offline.

    :param cpu: CPU number (e.g. 1, 2 or 39)
    :type cpu: int
    :return: `bool` True if online or False if not
    :rtype: bool
    """
    with open('/sys/devices/system/cpu/cpu%s/online' % cpu, 'rb') as cpu_online:
        if b'1' in cpu_online.read():
            return True
    return False


def cpu_has_flags(flags):
    """
    Check if a list of flags are available on current CPU info.

    :param flags: A `list` of cpu flags that must exists on the current CPU.
    :type flags: list of str
    :return: True if all the flags were found or False if not
    :rtype: bool
    """
    cpu_info = _get_info()

    if not isinstance(flags, list):
        flags = [flags]

    for flag in flags:
        if not any([flag.encode() in line for line in cpu_info]):
            return False
    return True


def get_version():
    """
    Get cpu version.

    :return: cpu version of given machine
             e.g.:- 'i5-5300U' for Intel and 'POWER9' for IBM machines in
             case of unknown/unsupported machines, return an empty string.
    :rtype: str
    """
    version_pattern = {'x86_64': rb'\s([\S,\d]+)\sCPU',
                       'i386': rb'\s([\S,\d]+)\sCPU',
                       'powerpc': rb'revision\s+:\s+(\S+)',
                       's390': rb'.*machine\s=\s(\d+)'
                       }
    cpu_info = _get_info()
    arch = get_arch()
    try:
        version_pattern[arch]
    except KeyError as Err:
        logging.warning("No pattern string for arch: %s\n Error: %s", arch, Err)
        return None
    for line in cpu_info:
        version_out = re.findall(version_pattern[arch], line)
        if version_out:
            return version_out[0].decode('utf-8')
    return ''


def get_vendor():
    """
    Get the current cpu vendor name.

    :return: a key of :data:`VENDORS_MAP` (e.g. 'intel') depending on the
             current CPU architecture. Return None if it was unable to
             determine the vendor name.
    :rtype: str or None
    """
    cpu_info = _get_info()
    for vendor, identifiers in VENDORS_MAP.items():
        for identifier in identifiers:
            if _list_matches(cpu_info, identifier):
                return vendor
    return None


def get_arch():
    """Work out which CPU architecture we're running on."""
    cpu_table = [(b'^cpu.*(RS64|Broadband Engine)', 'powerpc'),
                 (rb'^cpu.*POWER\d+', 'powerpc'),
                 (b'^cpu.*PPC970', 'powerpc'),
                 (b'(ARM|^CPU implementer|^CPU part|^CPU variant'
                  b'|^Features|^BogoMIPS|^CPU revision)', 'arm'),
                 (b'(^cpu MHz dynamic|^cpu MHz static|^features'
                  b'|^bogomips per cpu|^max thread id)', 's390'),
                 (b'^type', 'sparc64'),
                 (b'^flags.*:.* lm .*', 'x86_64'),
                 (b'^flags', 'i386'),
                 (b'^hart\\s*: 1$', 'riscv')]
    cpuinfo = _get_info()
    for (pattern, arch) in cpu_table:
        if _list_matches(cpuinfo, pattern):
            if arch == 'arm':
                # ARM is a special situation, which matches both 32 bits
                # (v7) and 64 bits (v8).
                for line in cpuinfo:
                    if line.startswith(b"CPU architecture"):
                        version = int(line.split(b':', 1)[1])
                        if version >= 8:
                            return 'aarch64'
                        else:
                            # For historical reasons return arm
                            return 'arm'
            return arch
    return platform.machine()


def get_family():
    """Get family name of the cpu like Broadwell, Haswell, power8, power9."""
    family = None
    arch = get_arch()
    if arch == 'x86_64' or arch == 'i386':
        if get_vendor() == 'amd':
            raise NotImplementedError
        try:
            # refer below links for microarchitectures names
            # https://en.wikipedia.org/wiki/List_of_Intel_CPU_microarchitectures
            # https://git.kernel.org/pub/scm/linux/kernel/git/torvalds/linux.git/tree/arch/x86/events/intel/core.c#n4613
            with open('/sys/devices/cpu/caps/pmu_name', 'rb') as mico_arch:
                family = mico_arch.read().decode('utf-8').strip('\n').lower()
        except FileNotFoundError as err:
            msg = "Could not find micro-architecture/family, Error: %s" % err
            logging.warning(msg)
            raise FamilyException(msg)
    elif arch == 'powerpc':
        res = []
        try:
            for line in _get_info():
                res = re.findall(rb'cpu\s+:\s+(POWER\d+)', line)
                if res:
                    break
            family = res[0].decode('utf-8').lower()
        except IndexError as err:
            msg = "Unable to parse cpu family %s" % err
            logging.warning(msg)
            raise FamilyException(msg)
    elif arch == 's390':
        zfamily_map = {'2964': 'z13',
                       '3906': 'z14',
                       '8561': 'z15'
                       }
        try:
            family = zfamily_map[get_version()].lower()
        except KeyError as err:
            msg = "Could not find family for %s\nError: %s" % (get_version(), err)
            logging.warning(msg)
            raise FamilyException(msg)
    else:
        raise NotImplementedError
    return family


def online_list():
    """Reports a list of indexes of the online cpus."""
    cpus = []
    search_str = b'processor'
    index = 2
    if platform.machine() == 's390x':
        search_str = b'cpu number'
        index = 3
    with open('/proc/cpuinfo', 'rb') as proc_cpuinfo:
        for line in proc_cpuinfo:
            if line.startswith(search_str):
                cpus.append(int(line.split()[index]))  # grab cpu number
    return cpus


def total_count():
    """Return Number of Total cpus in the system including offline cpus."""
    return os.sysconf('SC_NPROCESSORS_CONF')


def online_count():
    """Return Number of Online cpus in the system."""
    return os.sysconf('SC_NPROCESSORS_ONLN')


def online(cpu):
    """Online given CPU."""
    if _get_status(cpu) is False:
        with open("/sys/devices/system/cpu/cpu%s/online" % cpu, "wb") as fd:
            fd.write(b'1')
        if _get_status(cpu):
            return 0
    return 1


def offline(cpu):
    """Offline given CPU."""
    if _get_status(cpu):
        with open("/sys/devices/system/cpu/cpu%s/online" % cpu, "wb") as fd:
            fd.write(b'0')
        if _get_status(cpu):
            return 1
    return 0


def get_idle_state():
    """
    Get current cpu idle values.

    :return: Dict of cpuidle states values for all cpus
    :rtype: dict
    """
    cpus_list = online_list()
    states = len(glob.glob("/sys/devices/system/cpu/cpu0/cpuidle/state*"))
    cpu_idlestate = {}
    for cpu in cpus_list:
        cpu_idlestate[cpu] = {}
        for state_no in range(states):
            state_file = "/sys/devices/system/cpu/cpu%s/cpuidle/state%s/disable" % (cpu, state_no)
            try:
                cpu_idlestate[cpu][state_no] = bool(int(open(state_file, 'rb').read()))
            except IOError as err:
                logging.warning("Failed to read idle state on cpu %s "
                                "for state %s:\n%s", cpu, state_no, err)
    return cpu_idlestate


def _bool_to_binary(value):
    """
    Turns a Python boolean value (True or False) into binary data.

    This function is suitable for writing to /proc/* and /sys/* files.
    """
    if value is True:
        return b'1'
    if value is False:
        return b'0'
    raise TypeError("Value is not a boolean: %s" % value)


def set_idle_state(state_number="all", disable=True, setstate=None):
    """
    Set/Reset cpu idle states for all cpus.

    :param state_number: cpuidle state number, default: `all` all states
    :type state_number: str
    :param disable: whether to disable/enable given cpu idle state,
                    default is to disable.
    :type disable: bool
    :param setstate: cpuidle state value, output of `get_idle_state()`
    :type setstate: dict
    """
    cpus_list = online_list()
    if not setstate:
        states = []
        if state_number == 'all':
            states = list(range(len(glob.glob("/sys/devices/system/cpu/cpu0/cpuidle/state*"))))
        else:
            states.append(state_number)
        disable = _bool_to_binary(disable)
        for cpu in cpus_list:
            for state_no in states:
                state_file = "/sys/devices/system/cpu/cpu%s/cpuidle/state%s/disable" % (cpu, state_no)
                try:
                    open(state_file, "wb").write(disable)
                except IOError as err:
                    logging.warning("Failed to set idle state on cpu %s "
                                    "for state %s:\n%s", cpu, state_no, err)
    else:
        for cpu, stateval in setstate.items():
            for state_no, value in stateval.items():
                state_file = "/sys/devices/system/cpu/cpu%s/cpuidle/state%s/disable" % (cpu, state_no)
                disable = _bool_to_binary(value)
                try:
                    open(state_file, "wb").write(disable)
                except IOError as err:
                    logging.warning("Failed to set idle state on cpu %s "
                                    "for state %s:\n%s", cpu, state_no, err)


def set_freq_governor(governor="random"):
    """
    To change the given cpu frequency governor.

    :param governor: frequency governor profile name whereas `random` is default
                     option to choose random profile among available ones.
    :type governor: str
    """
    avl_gov_file = "/sys/devices/system/cpu/cpu0/cpufreq/scaling_available_governors"
    cur_gov_file = "/sys/devices/system/cpu/cpu0/cpufreq/scaling_governor"
    cur_gov = get_freq_governor()
    if not cur_gov:
        return False
    if not (os.access(avl_gov_file, os.R_OK) and os.access(cur_gov_file, os.W_OK)):
        logging.error("Could not locate frequency governor sysfs entries or\n"
                      " No proper permissions to read/write sysfs entries")
        return False
    cpus_list = total_count()
    with open(avl_gov_file, 'r') as fl:
        avl_govs = fl.read().strip().split(' ')
    if governor == "random":
        avl_govs.remove(cur_gov)
        if not avl_govs:
            logging.error("No other frequency governors to pick from...")
            return False
        governor = random.choice(avl_govs)
    if governor not in avl_govs:
        logging.warning("Trying to change unknown frequency "
                        "governor: %s", governor)
    for cpu in range(cpus_list):
        cur_gov_file = "/sys/devices/system/cpu/cpu%s/cpufreq/scaling_governor" % cpu
        try:
            with open(cur_gov_file, 'w') as fl:
                fl.write(governor)
        except IOError as err:
            logging.warning("Unable to write a given frequency "
                            "governor %s profile for cpu "
                            "%s\n %s", governor, cpu, err)
    return True


def get_freq_governor():
    """Get current cpu frequency governor."""
    cur_gov_file = "/sys/devices/system/cpu/cpu0/cpufreq/scaling_governor"
    try:
        with open(cur_gov_file, 'r') as fl:
            return fl.read().strip()
    except IOError as err:
        logging.error("Unable to get the current governor\n %s", err)
        return ""


def get_pid_cpus(pid):
    """
    Get all the cpus being used by the process according to pid informed.

    :param pid: process id
    :type pid: str
    :return: A list include all cpus the process is using
    :rtype: list
    """
    # processor id index is defined according proc documentation
    # the negative index is necessary because backward data
    # access has no misleading whitespaces
    processor_id_index = -14
    cpus = set()
    proc_stat_files = glob.glob('/proc/%s/task/[123456789]*/stat' % pid)

    for proc_stat_file in proc_stat_files:
        try:
            with open(proc_stat_file) as proc_stat:
                cpus.add(
                    proc_stat.read().split(' ')[processor_id_index]
                )
        except IOError:
            continue
    return list(cpus)


def _deprecated(newfunc, oldfuncname):
    """
    Print a warning to user and return the new function.

    :param newfunc: new function to be assigned
    :param oldfunctionname: Old function name string
    :rtype: `function`
    """
    def wrap(*args, **kwargs):
        fmt_str = "avocado.utils.cpu.{}() it is getting deprecat".format(oldfuncname)
        fmt_str += "ed, Use avocado.utils.cpu.{}() instead".format(newfunc.__name__)
        warnings.warn((fmt_str), DeprecationWarning, stacklevel=2)
        return newfunc(*args, **kwargs)
    return wrap


total_cpus_count = _deprecated(total_count, "total_cpus_count")
_get_cpu_info = _deprecated(_get_info, "_get_cpu_info")
_get_cpu_status = _deprecated(_get_status, "_get_cpu_status")
get_cpu_vendor_name = _deprecated(get_vendor, "get_cpu_vendor_name")
get_cpu_arch = _deprecated(get_arch, "get_cpu_arch")
cpu_online_list = _deprecated(online_list, "cpu_online_list")
online_cpus_count = _deprecated(online_count, "online_cpus_count")
get_cpuidle_state = _deprecated(get_idle_state, "get_cpuidle_state")
set_cpuidle_state = _deprecated(set_idle_state, "set_cpuidle_state")
set_cpufreq_governor = _deprecated(set_freq_governor, "set_cpufreq_governor")
get_cpufreq_governor = _deprecated(get_freq_governor, "get_cpufreq_governor")



from avocado import Test


class MyException(Exception):
    pass


def raises_exception():
    raise MyException


class Assert(Test):

    def test_assert_raises(self):
        with self.assertRaises(MyException):
            raises_exception()

    def test_fails_to_raise(self):
        with self.assertRaises(MyException):
            pass



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat Inc. 2013-2014,2016
# Author: Lucas Meneghel Rodrigues <lmr@redhat.com>
# Author: Lukas Doktor <ldoktor@redhat.com>

import json
import sys

from avocado.core import exit_codes
from avocado.core.output import LOG_UI
from avocado.core.plugin_interfaces import CLICmd
from avocado.core.settings import settings
from avocado.core.varianter import Varianter

_VERBOSITY_LEVELS = {"none": 0, "brief": 1, "normal": 2, "verbose": 3,
                     "full": 4, "max": 99}


def map_verbosity_level(level):
    if level.isdigit():
        return int(level)
    level = level.lower()
    if level in _VERBOSITY_LEVELS:
        return _VERBOSITY_LEVELS[level]
    else:
        raise ValueError


class Variants(CLICmd):

    """
    Implements "variants" command to visualize/debug test variants and params
    """

    name = 'variants'
    description = "Tool to analyze and visualize test variants and params"

    def configure(self, parser):
        parser = super(Variants, self).configure(parser)
        verbosity_levels = ("(positive integer - 0, 1, ... - or %s)"
                            % ", ".join(sorted(_VERBOSITY_LEVELS,
                                               key=lambda _: _VERBOSITY_LEVELS[_])))

        help_msg = 'Verbosity of the variants summary. ' + verbosity_levels
        settings.register_option(section='variants',
                                 key='summary',
                                 key_type=map_verbosity_level,
                                 default=0,
                                 help_msg=help_msg,
                                 parser=parser,
                                 long_arg='--summary')

        help_msg = 'Verbosity of the list of variants. ' + verbosity_levels
        settings.register_option(section='variants',
                                 key='variants',
                                 key_type=map_verbosity_level,
                                 default=1,
                                 help_msg=help_msg,
                                 parser=parser,
                                 long_arg='--variants')

        help_msg = ('[obsoleted by --variants] Shows the node content '
                    '(variables)')
        settings.register_option(section='variants',
                                 key='contents',
                                 key_type=bool,
                                 default=False,
                                 parser=parser,
                                 help_msg=help_msg,
                                 short_arg='-c',
                                 long_arg='--contents')

        help_msg = 'Dump the Variants to a JSON serialized file'
        settings.register_option(section='variants',
                                 key='json_variants_dump',
                                 help_msg=help_msg,
                                 default=None,
                                 parser=parser,
                                 long_arg='--json-variants-dump')

        env_parser = parser.add_argument_group("environment view options")

        help_msg = 'Use debug implementation to gather more information.'
        settings.register_option(section='variants',
                                 key='debug',
                                 help_msg=help_msg,
                                 parser=env_parser,
                                 key_type=bool,
                                 default=False,
                                 long_arg='--debug',
                                 short_arg='-d')

        tree_parser = parser.add_argument_group("tree view options")

        help_msg = ('[obsoleted by --summary] Shows the multiplex tree '
                    'structure')
        settings.register_option(section='variants',
                                 key='tree',
                                 key_type=bool,
                                 default=False,
                                 help_msg=help_msg,
                                 parser=tree_parser,
                                 long_arg='--tree',
                                 short_arg='-t')

        help_msg = '[obsoleted by --summary] Show the inherited values'
        settings.register_option(section='variants',
                                 key='inherit',
                                 key_type=bool,
                                 help_msg=help_msg,
                                 parser=tree_parser,
                                 default=False,
                                 short_arg='-i',
                                 long_arg='--inherit')

    def run(self, config):
        tree = config.get('variants.tree')
        summary = config.get('variants.summary')
        variants = config.get('variants.variants')
        contents = config.get('variants.contents')
        inherit = config.get('variants.inherit')

        err = None
        if tree and config.get('variants.debug'):
            err = "Option --tree is incompatible with --debug."
        elif not tree and inherit:
            err = "Option --inherit can be only used with --tree"
        if err:
            LOG_UI.error(err)
            sys.exit(exit_codes.AVOCADO_FAIL)
        varianter = Varianter()
        try:
            varianter.parse(config)
        except (IOError, ValueError) as details:
            LOG_UI.error("Unable to parse varianter: %s", details)
            sys.exit(exit_codes.AVOCADO_FAIL)
        use_utf8 = config.get("runner.output.utf8")
        # Parse obsolete options (unsafe to combine them with new args)
        if tree:
            variants = 0
            summary += 1
            if contents:
                summary += 1
            if inherit:
                summary += 2
        else:
            if contents:
                variants += 2

        json_variants_dump = config.get('variants.json_variants_dump')
        # Export the serialized variants
        if json_variants_dump is not None:
            try:
                with open(json_variants_dump, 'w') as variants_file:
                    json.dump(varianter.dump(), variants_file)
            except IOError:
                LOG_UI.error("Cannot write %s", json_variants_dump)
                sys.exit(exit_codes.AVOCADO_FAIL)

        # Produce the output
        lines = varianter.to_str(summary=summary,
                                 variants=variants,
                                 use_utf8=use_utf8)
        for line in lines.splitlines():
            LOG_UI.debug(line)

        sys.exit(exit_codes.AVOCADO_ALL_OK)



import time

from avocado.core.output import LOG_UI
from avocado.core.plugin_interfaces import Init, JobPost, JobPre
from avocado.core.settings import settings


class SleepInit(Init):
    name = 'sleep-init'
    description = 'Sleep plugin initialization'

    def initialize(self):
        help_msg = 'Number of seconds to sleep.'
        settings.register_option(section='plugins.job.sleep',
                                 key='seconds',
                                 default=3,
                                 help_msg=help_msg)


class Sleep(JobPre, JobPost):
    name = 'sleep'
    description = 'Sleeps for a number of seconds'

    def sleep(self, job):  # pylint: disable=W0613
        seconds = job.config.get('plugins.job.sleep.seconds')
        for i in range(1, seconds + 1):
            LOG_UI.info("Sleeping %2i/%s", i, seconds)
            time.sleep(1)

    pre = post = sleep



import os
import tempfile
import unittest

from avocado.utils.filelock import AlreadyLocked, FileLock

from .. import temp_dir_prefix


class TestFileLock(unittest.TestCase):

    def setUp(self):
        prefix = temp_dir_prefix(__name__, self, 'setUp')
        self.tmpdir = tempfile.TemporaryDirectory(prefix=prefix)
        self.filename = os.path.join(self.tmpdir.name, 'file.img')
        self.content = 'Foo bar'
        with open(self.filename, 'w') as f:
            f.write(self.content)

    def _readfile(self):
        with FileLock(self.filename):
            with open(self.filename, 'r') as f:
                return f.read()

    def test_readfile(self):
        self.assertEqual(self._readfile(), self.content)

    def test_locked_by_me(self):
        with FileLock(self.filename):
            self.assertRaises(AlreadyLocked, self._readfile)

    def test_locked_by_other(self):
        with open(self.filename+'.lock', 'w') as f:
            f.write('1')
        self.assertRaises(AlreadyLocked, self._readfile)

    def tearDown(self):
        self.tmpdir.cleanup()


if __name__ == "__main__":
    unittest.main()



import enum
import re
from collections import namedtuple


@enum.unique
class TestResult(enum.Enum):
    PASS = 'PASS'
    SKIP = 'SKIP'
    FAIL = 'FAIL'
    XFAIL = 'XFAIL'
    XPASS = 'XPASS'


# TapParser is based on Meson's TAP parser, which were licensed under the
# MIT (X11) license and were contributed to both Meson and Avocado by the
# same author (Paolo).

class TapParser:
    Plan = namedtuple('Plan', ['count', 'late', 'skipped', 'explanation'])
    Bailout = namedtuple('Bailout', ['message'])
    Test = namedtuple('Test', ['number', 'name', 'result', 'explanation'])
    Error = namedtuple('Error', ['message'])
    Version = namedtuple('Version', ['version'])

    _MAIN = 1
    _AFTER_TEST = 2
    _YAML = 3

    _RE_BAILOUT = re.compile(r'Bail out!\s*(.*)')
    _RE_DIRECTIVE = re.compile(r'(?:\s*\#\s*([Ss][Kk][Ii][Pp]\S*|[Tt][Oo][Dd][Oo])\b\s*(.*))?')
    _RE_PLAN = re.compile(r'1\.\.([0-9]+)' + _RE_DIRECTIVE.pattern)
    _RE_TEST = re.compile(r'((?:not )?ok)\s*(?:([0-9]+)\s*)?([^#]*)' + _RE_DIRECTIVE.pattern)
    _RE_VERSION = re.compile(r'TAP version ([0-9]+)')
    _RE_YAML_START = re.compile(r'(\s+)---.*')
    _RE_YAML_END = re.compile(r'\s+\.\.\.\s*')

    def __init__(self, tap_io):
        self.tap_io = tap_io

    def parse_test(self, ok, num, name, directive, explanation):
        name = name.strip()
        explanation = explanation.strip() if explanation else None
        if directive is not None:
            directive = directive.upper()
            if directive == 'SKIP':
                if ok:
                    yield self.Test(num, name, TestResult.SKIP, explanation)
                    return
            elif directive == 'TODO':
                result = TestResult.XPASS if ok else TestResult.XFAIL
                yield self.Test(num, name, result, explanation)
                return
            else:
                yield self.Error('invalid directive "%s"' % (directive,))

        result = TestResult.PASS if ok else TestResult.FAIL
        yield self.Test(num, name, result, explanation)

    def parse(self):
        found_late_test = False
        bailed_out = False
        plan = None
        lineno = 0
        num_tests = 0
        yaml_lineno = 0
        yaml_indent = ''
        state = self._MAIN
        version = 12
        while True:
            lineno += 1
            try:
                line = next(self.tap_io).rstrip()
            except StopIteration:
                break

            # YAML blocks are only accepted after a test
            if state == self._AFTER_TEST:
                if version >= 13:
                    m = self._RE_YAML_START.match(line)
                    if m:
                        state = self._YAML
                        yaml_lineno = lineno
                        yaml_indent = m.group(1)
                        continue
                state = self._MAIN

            elif state == self._YAML:
                if self._RE_YAML_END.match(line):
                    state = self._MAIN
                    continue
                if line.startswith(yaml_indent):
                    continue
                yield self.Error('YAML block not terminated (started on line %d)' % (yaml_lineno,))
                state = self._MAIN

            assert state == self._MAIN
            if line.startswith('#'):
                continue

            m = self._RE_TEST.match(line)
            if m:
                if plan and plan.late and not found_late_test:
                    yield self.Error('unexpected test after late plan')
                    found_late_test = True
                num_tests += 1
                num = num_tests if m.group(2) is None else int(m.group(2))
                if num != num_tests:
                    yield self.Error('out of order test numbers')
                yield from self.parse_test(m.group(1) == 'ok', num,
                                           m.group(3), m.group(4), m.group(5))
                state = self._AFTER_TEST
                continue

            m = self._RE_PLAN.match(line)
            if m:
                if plan:
                    yield self.Error('more than one plan found')
                else:
                    count = int(m.group(1))
                    skipped = (count == 0)
                    if m.group(2):
                        if m.group(2).upper().startswith('SKIP'):
                            if count > 0:
                                yield self.Error('invalid SKIP directive for plan')
                            skipped = True
                        else:
                            yield self.Error('invalid directive for plan')
                    plan = self.Plan(count=count, late=(num_tests > 0),
                                     skipped=skipped, explanation=m.group(3))
                    yield plan
                continue

            m = self._RE_BAILOUT.match(line)
            if m:
                yield self.Bailout(m.group(1))
                bailed_out = True
                continue

            m = self._RE_VERSION.match(line)
            if m:
                # The TAP version is only accepted as the first line
                if lineno != 1:
                    yield self.Error('version number must be on the first line')
                    continue
                version = int(m.group(1))
                if version < 13:
                    yield self.Error('version number should be at least 13')
                else:
                    yield self.Version(version=version)
                continue

            if line == '':
                continue

            yield self.Error('unexpected input at line %d' % (lineno,))

        if state == self._YAML:
            yield self.Error('YAML block not terminated (started on line %d)' % (yaml_lineno,))

        if not bailed_out and plan and num_tests != plan.count:
            if num_tests < plan.count:
                yield self.Error('Too few tests run (expected %d, got %d)'
                                 % (plan.count, num_tests))
            else:
                yield self.Error('Too many tests run (expected %d, got %d)'
                                 % (plan.count, num_tests))



from setuptools import setup

if __name__ == '__main__':
    setup(name='avocado-hello-world-option',
          version='1.0',
          description='Avocado Hello World CLI command with config option',
          py_modules=['hello_option'],
          entry_points={
              'avocado.plugins.cli.cmd': ['hello_option = hello_option:HelloWorld'],
              }
          )



#
# Copyright (c) 2008 Michael Eddington
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Authors:
#   Frank Laub (frank.laub@gmail.com)
#   Michael Eddington (mike@phed.org)


import pprint
import re

from . import spark


class Token:
    def __init__(self, token_type, value=None):
        self.type = token_type
        self.value = value

    def __lt__(self, o):
        return self.type < o

    def __gt__(self, o):
        return self.type > o

    def __le__(self, o):
        return self.type <= o

    def __ge__(self, o):
        return self.type >= o

    def __eq__(self, o):
        return self.type == o

    def __ne__(self, o):
        return self.type != o

    def __repr__(self):
        return self.value or self.type


class AST:
    def __init__(self, ast_type):
        self.type = ast_type
        self._kids = []

    def __getitem__(self, i):
        return self._kids[i]

    def __setitem__(self, i, k):
        self._kids[i] = k

    def __len__(self):
        return len(self._kids)

    def __lt__(self, o):
        return self.type < o

    def __gt__(self, o):
        return self.type > o

    def __le__(self, o):
        return self.type <= o

    def __ge__(self, o):
        return self.type >= o

    def __eq__(self, o):
        return self.type == o

    def __ne__(self, o):
        return self.type != o


class GdbMiScannerBase(spark.GenericScanner):
    def tokenize(self, data_input):  # pylint: disable=W0221
        self.rv = []  # pylint: disable=W0201
        spark.GenericScanner.tokenize(self, data_input)
        return self.rv

    def t_nl(self, s):  # pylint: disable=W0613
        r'\n|\r\n'
        self.rv.append(Token('nl'))

    def t_whitespace(self, s):  # pylint: disable=W0613
        r'[ \t\f\v]+'

    def t_symbol(self, s):
        r',|\{|\}|\[|\]|\='
        self.rv.append(Token(s, s))

    def t_result_type(self, s):
        r'\*|\+|\^'
        self.rv.append(Token('result_type', s))

    def t_stream_type(self, s):
        r'\@|\&|\~'
        self.rv.append(Token('stream_type', s))

    def t_string(self, s):
        r'[\w-]+'
        self.rv.append(Token('string', s))

    def t_c_string(self, s):
        r'\".*?(?<![\\\\])\"'
        inner = self.__unescape(s[1:len(s)-1])
        self.rv.append(Token('c_string', inner))

    def t_default(self, s):
        r'( . | \n )+'
        raise Exception("Specification error: unmatched input for '%s'" % s)

    def __unescape(self, s):
        s = re.sub(r'\\r', r'\r', s)
        s = re.sub(r'\\n', r'\n', s)
        s = re.sub(r'\\t', r'\t', s)
        return re.sub(r'\\(.)', r'\1', s)


class GdbMiScanner(GdbMiScannerBase):
    def t_token(self, s):
        r'\d+'
        self.rv.append(Token('token', s))


class GdbMiParser(spark.GenericASTBuilder):
    def __init__(self):
        spark.GenericASTBuilder.__init__(self, AST, 'output')

    def p_output(self, args):
        '''
                output ::= record_list
                record_list ::= generic_record
                record_list ::= generic_record record_list
                generic_record ::= result_record
                generic_record ::= stream_record
                result_record ::= result_header result_list nl
                result_record ::= result_header nl
                result_header ::= token result_type class
                result_header ::= result_type class
                result_header ::= token = class
                result_header ::= = class
                stream_record ::= stream_type c_string nl
                result_list ::= , result result_list
                result_list ::= , result
                result_list ::= , tuple
                result ::= variable = value
                class ::= string
                variable ::= string
                value ::= const
                value ::= tuple
                value ::= list
                value_list ::= , value
                value_list ::= , value value_list
                const ::= c_string
                tuple ::= { }
                tuple ::= { result }
                tuple ::= { result result_list }
                list ::= [ ]
                list ::= [ value ]
                list ::= [ value value_list ]
                list ::= [ result ]
                list ::= [ result result_list ]
                list ::= { value }
                list ::= { value value_list }
        '''

    def terminal(self, token):
        #  Homogeneous AST.
        rv = AST(token.type)
        rv.value = token.value  # pylint: disable=W0201
        return rv

    def nonterminal(self, token_type, args):
        #  Flatten AST a bit by not making nodes if there's only one child.
        exclude = [
            'record_list'
        ]
        if len(args) == 1 and token_type not in exclude:
            return args[0]
        return spark.GenericASTBuilder.nonterminal(self, token_type, args)

    def error(self, token, i=0, tokens=None):  # pylint: disable=W0221
        if i > 2:
            print('%s %s %s %s' % (tokens[i-3], tokens[i-2], tokens[i-1], tokens[i]))
        raise Exception("Syntax error at or near %d:'%s' token" % (i, token))


class GdbMiInterpreter(spark.GenericASTTraversal):
    def __init__(self, ast):
        spark.GenericASTTraversal.__init__(self, ast)
        self.postorder()

    def __translate_type(self, token_type):
        table = {
            '^': 'result',
            '=': 'notify',
            '+': 'status',
            '*': 'exec',
            '~': 'console',
            '@': 'target',
            '&': 'log'
        }
        return table[token_type]

    def n_result(self, node):
        # result ::= variable = value
        node.value = {node[0].value: node[2].value}
        # print 'result: %s' % node.value

    def n_tuple(self, node):
        if len(node) == 2:
            # tuple ::= {}
            node.value = {}
        elif len(node) == 3:
            # tuple ::= { result }
            node.value = node[1].value
        elif len(node) == 4:
            # tuple ::= { result result_list }
            node.value = node[1].value
            for result in node[2].value:
                for n, v in list(result.items()):
                    if n in node.value:
                        # print '**********list conversion: [%s] %s -> %s' % (n, node.value[n], v)
                        old = node.value[n]
                        if not isinstance(old, list):
                            node.value[n] = [node.value[n]]
                        node.value[n].append(v)
                    else:
                        node.value[n] = v
        else:
            raise Exception('Invalid tuple')
        # print 'tuple: %s' % node.value

    def n_list(self, node):
        if len(node) == 2:
            # list ::= []
            node.value = []
        elif len(node) == 3:
            # list ::= [ value ]
            node.value = [node[1].value]
        elif len(node) == 4:
            # list ::= [ value value_list ]
            node.value = [node[1].value] + node[2].value
            # list ::= [ result ]
            # list ::= [ result result_list ]
            # list ::= { value }
            # list ::= { value value_list }
        # print 'list %s' % node.value

    def n_value_list(self, node):
        if len(node) == 2:
            # value_list ::= , value
            node.value = [node[1].value]
        elif len(node) == 3:
            # value_list ::= , value value_list
            node.value = [node[1].value] + node[2].value

    def n_result_list(self, node):
        if len(node) == 2:
            # result_list ::= , result
            node.value = [node[1].value]
        else:
            # result_list ::= , result result_list
            node.value = [node[1].value] + node[2].value
        # print 'result_list: %s' % node.value

    def n_result_record(self, node):
        node.value = node[0].value
        if len(node) == 3:
            # result_record ::= result_header result_list nl
            node.value['results'] = node[1].value
        elif len(node) == 2:
            # result_record ::= result_header nl
            pass
        # print 'result_record: %s' % (node.value)

    def n_result_header(self, node):
        if len(node) == 3:
            # result_header ::= token result_type class
            node.value = {
                    'token': node[0].value,
                    'type': self.__translate_type(node[1].value),
                    'class_': node[2].value,
                    'record_type': 'result'
            }
        elif len(node) == 2:
            # result_header ::= result_type class
            node.value = {
                    'token': None,
                    'type': self.__translate_type(node[0].value),
                    'class_': node[1].value,
                    'record_type': 'result'
            }

    def n_stream_record(self, node):
        # stream_record ::= stream_type c_string nl
        node.value = {
            'type': self.__translate_type(node[0].value),
            'value': node[1].value,
            'record_type': 'stream'
        }
        # print 'stream_record: %s' % node.value

    def n_record_list(self, node):
        if len(node) == 1:
            # record_list ::= generic_record
            node.value = [node[0].value]
        elif len(node) == 2:
            # record_list ::= generic_record record_list
            node.value = [node[0].value] + node[1].value
        # print 'record_list: %s' % node.value

    # def default(self, node):
        # print 'default: ' + node.type


class GdbDynamicObject:
    def __init__(self, dict_):
        self.graft(dict_)

    def __repr__(self):
        return pprint.pformat(self.__dict__)

    def __bool__(self):
        return len(self.__dict__) > 0

    def __getitem__(self, i):
        if i == 0 and len(self.__dict__) > 0:
            return self
        else:
            raise IndexError

    def __getattr__(self, name):
        if name.startswith('__'):
            raise AttributeError
        return None

    def graft(self, dict_):
        for name, value in list(dict_.items()):
            name = name.replace('-', '_')
            if isinstance(value, dict):
                value = GdbDynamicObject(value)
            elif isinstance(value, list):
                x = value
                value = []
                for item in x:
                    if isinstance(item, dict):
                        item = GdbDynamicObject(item)
                    value.append(item)
            setattr(self, name, value)


class GdbMiRecord:
    def __init__(self, record):
        self.result = None
        for name, value in list(record[0].items()):
            name = name.replace('-', '_')
            if name == 'results':
                for result in value:
                    if not self.result:
                        self.result = GdbDynamicObject(result)
                    else:
                        # graft this result to self.results
                        self.result.graft(result)
            else:
                setattr(self, name, value)

    def __repr__(self):
        return pprint.pformat(self.__dict__)


class session:
    def __init__(self):
        self.the_scanner = GdbMiScanner()
        self.the_parser = GdbMiParser()
        self.the_interpreter = GdbMiInterpreter
        self.the_output = GdbMiRecord

    def scan(self, data_input):
        return self.the_scanner.tokenize(data_input)

    def parse(self, tokens):
        return self.the_parser.parse(tokens)

    def process(self, data_input):
        tokens = self.scan(data_input)
        ast = self.parse(tokens)
        self.the_interpreter(ast)
        return self.the_output(ast.value)



from avocado import Test


class GetData(Test):

    """
    Example for get_data() API usage
    """

    def test_a(self):
        """
        This large (on purpose) test, tests get_data() with "file",
        "test" and "variant" sources.

        Then, it adds other checks that include all sources.
        """
        # File-level checks
        file_data = self.get_data('file_data')
        self.assertIsNotNone(file_data)
        self.assertEqual(file_data,
                         self.get_data('file_data', source='file'))
        self.assertEqual(file_data,
                         self.get_data('file_data', source='file',
                                       must_exist=False))
        self.assertEqual(open(file_data).read(), 'get_data.py')

        # Test-level checks
        test_data = self.get_data('test_data')
        self.assertIsNotNone(test_data)
        self.assertEqual(test_data,
                         self.get_data('test_data', source='test'))
        self.assertEqual(test_data,
                         self.get_data('test_data', source='test',
                                       must_exist=False))
        self.assertEqual(open(test_data).read(), 'a')

        # Variant-level checks
        in_variant = self.params.get('in_variant', default=False)
        if in_variant:
            variant_data = self.get_data('variant_data')
            self.assertIsNotNone(variant_data)
            self.assertEqual(variant_data,
                             self.get_data('variant_data', source='variant'))
            self.assertEqual(variant_data,
                             self.get_data('variant_data', source='variant',
                                           must_exist=False))

        # A variation of data files that do not exist
        self.assertIsNone(self.get_data('does_not_exist'))
        self.assertIsNone(self.get_data('file_data', source='test'))
        self.assertIsNone(self.get_data('test_data', source='file'))
        if in_variant:
            self.assertIsNone(self.get_data('variant_data', source='test'))

        # All `get_data()` called with `must_exist=False` should
        # return a valid location for a (to be created?) data file
        self.assertIsNotNone(self.get_data('does_not_exist', must_exist=False))
        self.assertIsNotNone(self.get_data('does_not_exist', source='file',
                                           must_exist=False))
        self.assertIsNotNone(self.get_data('does_not_exist', source='test',
                                           must_exist=False))
        if in_variant:
            self.assertIsNotNone(self.get_data('does_not_exist',
                                               source='variant',
                                               must_exist=False))

        # Write to stdout with print() to test output check capabilities,
        # a feature that uses the data directories (get_data()) itself
        print("This is output from test_a")

    def test_b(self):
        """
        Test contains data files under the test data directory, but
        has no associated variants directories.
        """
        file_data = self.get_data('file_data')
        self.assertIsNotNone(file_data)
        self.assertEqual(open(file_data).read(), 'get_data.py')

        test_data = self.get_data('test_data')
        self.assertIsNotNone(test_data)
        self.assertEqual(open(test_data).read(), 'b')

        variant_data = self.get_data('variant_data')
        self.assertIsNone(variant_data)

        print("This is output from test_b")



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat Inc. 2013-2014
# Author: Lucas Meneghel Rodrigues <lmr@redhat.com>


import os

import pkg_resources

from .dispatcher import InitDispatcher
from .settings import settings as stgs
from .streams import BUILTIN_STREAM_SETS, BUILTIN_STREAMS
from .utils import prepend_base_path


def register_core_options():
    streams = (['"%s": %s' % _ for _ in BUILTIN_STREAMS.items()] +
               ['"%s": %s' % _ for _ in BUILTIN_STREAM_SETS.items()])
    streams = "; ".join(streams)
    help_msg = ("List of comma separated builtin logs, or logging streams "
                "optionally followed by LEVEL (DEBUG,INFO,...). Builtin "
                "streams are: %s. By default: 'app'" % streams)
    stgs.register_option(section='core',
                         key='show',
                         key_type=lambda x: x.split(','),
                         metavar="STREAM[:LVL]",
                         nargs='?',
                         default=['app'],
                         help_msg=help_msg)

    help_msg = ('Python regular expression that will make the test '
                'status WARN when matched.')
    stgs.register_option(section='simpletests.status',
                         key='warn_regex',
                         default='^WARN$',
                         help_msg=help_msg)

    help_msg = ('Location to search the regular expression on. '
                'Accepted values: all, stdout, stderr.')
    stgs.register_option(section='simpletests.status',
                         key='warn_location',
                         default='all',
                         help_msg=help_msg)

    help_msg = ('Python regular expression that will make the test '
                'status SKIP when matched.')
    stgs.register_option(section='simpletests.status',
                         key='skip_regex',
                         default='^SKIP$',
                         help_msg=help_msg)

    help_msg = ('Location to search the regular expression on. '
                'Accepted values: all, stdout, stderr.')
    stgs.register_option(section='simpletests.status',
                         key='skip_location',
                         default='all',
                         help_msg=help_msg)

    help_msg = ('The amount of time to give to the test process after '
                'it it has been interrupted (such as with CTRL+C)')
    stgs.register_option(section='runner.timeout',
                         key='after_interrupted',
                         key_type=int,
                         help_msg=help_msg,
                         default=60)

    help_msg = 'Cache directories to be used by the avocado test'
    stgs.register_option(section='datadir.paths',
                         key='cache_dirs',
                         key_type=list,
                         default=[],
                         help_msg=help_msg)

    help_msg = 'Base directory for Avocado tests and auxiliary data'
    default = prepend_base_path('/var/lib/avocado')
    stgs.register_option(section='datadir.paths',
                         key='base_dir',
                         key_type=prepend_base_path,
                         default=default,
                         help_msg=help_msg)

    help_msg = 'Test directory for Avocado tests'
    default = prepend_base_path('/usr/share/doc/avocado/tests')
    stgs.register_option(section='datadir.paths',
                         key='test_dir',
                         key_type=prepend_base_path,
                         default=default,
                         help_msg=help_msg)

    help_msg = 'Data directory for Avocado'
    default = prepend_base_path('/var/lib/avocado/data')
    stgs.register_option(section='datadir.paths',
                         key='data_dir',
                         key_type=prepend_base_path,
                         default=default,
                         help_msg=help_msg)

    help_msg = 'Logs directory for Avocado'
    default = prepend_base_path('~/avocado/job-results')
    stgs.register_option(section='datadir.paths',
                         key='logs_dir',
                         key_type=prepend_base_path,
                         default=default,
                         help_msg=help_msg)

    help_msg = ('The amount of time to wait after a test has reported '
                'status but the test process has not finished')
    stgs.register_option(section='runner.timeout',
                         key='process_alive',
                         key_type=int,
                         help_msg=help_msg,
                         default=60)

    help_msg = ('The amount of to wait for a test status after the '
                'process has been noticed to be dead')
    stgs.register_option(section='runner.timeout',
                         key='process_died',
                         key_type=int,
                         help_msg=help_msg,
                         default=10)
    help_msg = ('Whether to display colored output in terminals that '
                'support it')
    stgs.register_option(section='runner.output',
                         key='colored',
                         key_type=bool,
                         default=True,
                         help_msg=help_msg)

    help_msg = ('Whether to force colored output to non-tty outputs '
                '(e.g. log files). Allowed values: auto, always, never')
    stgs.register_option(section='runner.output',
                         key='color',
                         default='auto',
                         help_msg=help_msg)

    help_msg = 'Use UTF8 encoding (True or False)'
    stgs.register_option(section='runner.output',
                         key='utf8',
                         key_type=bool,
                         help_msg=help_msg,
                         default=True)

    help_msg = ('Suppress notification about broken plugins in the app '
                'standard error. Add the name of each broken plugin you '
                'want to suppress the notification in the list. (e.g. '
                '"avocado_result_html")')
    stgs.register_option(section='plugins',
                         key='skip_broken_plugin_notification',
                         key_type=list,
                         default=[],
                         help_msg=help_msg)

    help_msg = 'The encoding used by default on all data input'
    stgs.register_option(section='core',
                         key='input_encoding',
                         default='utf-8',
                         help_msg=help_msg)

    # All settings starting with 'runner.' will be passed to runner
    # exec-test runner config
    help_msg = ('Use a custom exit code list to consider a test as skipped. '
                'This is only used by exec-test runners. Default is [].')
    stgs.register_option(section='runner.exectest.exitcodes',
                         key='skip',
                         default=[],
                         key_type=list,
                         help_msg=help_msg)


def initialize_plugin_infrastructure():
    help_msg = 'Plugins that will not be loaded and executed'
    stgs.register_option(section='plugins',
                         key='disable',
                         key_type=list,
                         default=[],
                         help_msg=help_msg)

    kinds = pkg_resources.get_entry_map('avocado-framework').keys()
    plugin_types = [kind[8:] for kind in kinds
                    if kind.startswith('avocado.plugins.')]
    for plugin_type in plugin_types:
        help_msg = 'Execution order for "%s" plugins' % plugin_type
        stgs.register_option(section=plugin_type,
                             key='order',
                             key_type=list,
                             default=[],
                             help_msg=help_msg)


def initialize_plugins():
    initialize_plugin_infrastructure()
    InitDispatcher().map_method('initialize')



import unittest

from avocado.utils import data_structures


class TestDataStructures(unittest.TestCase):

    """
    Unit tests for avocado.utils.data_structures
    """

    def test_ordered_list_unique(self):
        """
        Verify only unique items are there and the order is kept
        """
        orig = [1, 5, unittest, 1, 3, 'a', 7, 7, None, 0, unittest]
        self.assertEqual(data_structures.ordered_list_unique(orig),
                         [1, 5, unittest, 3, 'a', 7, None, 0])

    def test_geometric_mean(self):
        """
        Verify the correct value is produced and it allows processing of long
        lists of values where some algorithm fails.
        """
        self.assertEqual(data_structures.geometric_mean(list(range(1, 180))),
                         67.1555819421869)

    def test_compare_matrices(self):
        """
        Verify the correct value is produced when comparing matrices.
        """
        # Note that first row contains header in first column, while the
        # second contains only values (for testing purposes)
        matrix1 = [["header", 51.7, 60], [1, 0, 0]]
        matrix2 = [["header", 57.2, 54], [2, 51, 0]]
        self.assertEqual(data_structures.compare_matrices(matrix1, matrix2),
                         ([["header", '+10.6383', -10.0],
                           ['+100', 'error_51/0', '.']], 3, 1, 5))

    def test_comma_separated_ranges_to_list(self):
        """
        Verify the correct value is obtained when converting a comma separated
        range string to list
        """
        node_values = ["0", "1-3", "0-1,16-17", "0-1,16-20,23-25"]
        expected_values = [[0], [1, 2, 3], [0, 1, 16, 17],
                           [0, 1, 16, 17, 18, 19, 20, 23, 24, 25]]
        for index, value in enumerate(node_values):
            self.assertEqual(data_structures.comma_separated_ranges_to_list(
                value), expected_values[index])

    def test_lazy_property(self):
        """
        Verify the value is initialized lazily with the correct value
        """
        class DummyClass:
            value = False

            @data_structures.LazyProperty
            def dummy_method(self):
                return not self.value

        item = DummyClass()
        self.assertNotIn('dummy_method', item.__dict__)
        self.assertEqual(item.dummy_method, True)
        self.assertIn('dummy_method', item.__dict__)

    def test_callback_register(self):
        """
        Checks CallbackRegister
        """
        class Log:
            msgs = []

            def error(self, *args, **kwargs):
                self.msgs.append((args, kwargs))

        def ret_arg(arg):
            return arg

        log = Log()
        register = data_structures.CallbackRegister("MyName", log)
        # Register few correct functions
        register.register(ret_arg, [True], {})
        register.register(ret_arg, [], {"arg": False})
        # Register few incorrect functions (incorrect number of arguments)
        register.register(ret_arg, [True, "incorrect_twice"], {})
        register.register(ret_arg, [True, "incorrect_twice"], {})
        register.register(ret_arg, [True, "incorrect_unique_only"], {}, True)
        # Register incorrect function, which is removed before cleanup
        register.register(ret_arg, [True, "incorrect_removed"], {})
        register.unregister(ret_arg, [True, "incorrect_removed"], {})
        # Run registered functions and check errors were produced accordingly
        register.run()
        self.assertEqual(len(log.msgs), 3)
        self.assertIn("incorrect_unique_only", str(log.msgs[0]))
        self.assertIn("incorrect_twice", str(log.msgs[1]))
        self.assertIn("incorrect_twice", str(log.msgs[2]))

    def test_time_to_seconds(self):
        self.assertEqual(data_structures.time_to_seconds(None), 0)
        self.assertEqual(data_structures.time_to_seconds("31"), 31)
        self.assertEqual(data_structures.time_to_seconds('10d'), 864000)
        self.assertRaises(ValueError, data_structures.time_to_seconds,
                          "10days")


class TestDataSize(unittest.TestCase):

    def test_valid(self):
        data_structures.DataSize('0')
        data_structures.DataSize('0t')
        data_structures.DataSize('10')

    def test_invalid(self):
        self.assertRaises(data_structures.InvalidDataSize,
                          data_structures.DataSize, 'megabyte')
        self.assertRaises(data_structures.InvalidDataSize,
                          data_structures.DataSize, '-100t')
        self.assertRaises(data_structures.InvalidDataSize,
                          data_structures.DataSize, '0.5g')
        self.assertRaises(data_structures.InvalidDataSize,
                          data_structures.DataSize, '10Mb')

    def test_value_and_type(self):
        self.assertEqual(data_structures.DataSize('0b').b, 0)
        self.assertEqual(data_structures.DataSize('0t').b, 0)

    def test_values(self):
        self.assertEqual(data_structures.DataSize('10m').b, 10485760)
        self.assertEqual(data_structures.DataSize('10M').b, 10485760)


if __name__ == "__main__":
    unittest.main()



import unittest

from avocado.core.test_id import TestID
from avocado.utils import astring


class Test(unittest.TestCase):

    def test_uid_name(self):
        uid = 1
        name = 'file.py:klass.test_method'
        test_id = TestID(uid, name)
        self.assertEqual(test_id.uid, 1)
        self.assertEqual(test_id.str_uid, '1')
        self.assertEqual(test_id.str_filesystem,
                         astring.string_to_safe_path('%s-%s' % (uid, name)))
        self.assertIs(test_id.variant, None)
        self.assertIs(test_id.str_variant, '')

    def test_uid_name_no_digits(self):
        uid = 1
        name = 'file.py:klass.test_method'
        test_id = TestID(uid, name, no_digits=2)
        self.assertEqual(test_id.uid, 1)
        self.assertEqual(test_id.str_uid, '01')
        self.assertEqual(test_id.str_filesystem,
                         astring.string_to_safe_path('%s-%s' % ('01', name)))
        self.assertIs(test_id.variant, None)
        self.assertIs(test_id.str_variant, '')

    def test_uid_name_large_digits(self):
        """
        Tests that when the filesystem can only cope with the size of
        the Test ID, that's the only thing that will be kept.
        """
        uid = 1
        name = 'test'
        test_id = TestID(uid, name, no_digits=255)
        self.assertEqual(test_id.uid, 1)
        self.assertEqual(test_id.str_uid, '%0255i' % uid)
        self.assertEqual(test_id.str_filesystem, '%0255i' % uid)
        self.assertIs(test_id.variant, None)
        self.assertIs(test_id.str_variant, '')

    def test_uid_name_uid_too_large_digits(self):
        """
        Tests that when the filesystem can not cope with the size of
        the Test ID, not even the test uid, an exception will be
        raised.
        """
        test_id = TestID(1, 'test', no_digits=256)
        self.assertRaises(RuntimeError, lambda: test_id.str_filesystem)

    def test_uid_large_name(self):
        """
        Tests that when the filesystem can not cope with the size of
        the Test ID, the name will be shortened.
        """
        uid = 1
        name = 'test_' * 51     # 255 characters
        test_id = TestID(uid, name)
        self.assertEqual(test_id.uid, 1)
        # only 253 can fit for the test name
        self.assertEqual(test_id.str_filesystem, '%s-%s' % (uid, name[:253]))
        self.assertIs(test_id.variant, None)
        self.assertIs(test_id.str_variant, "")

    def test_uid_name_large_variant(self):
        """
        Tests that when the filesystem can not cope with the size of
        the Test ID, and a variant name is present, the name will be
        removed.
        """
        uid = 1
        name = 'test'
        variant_id = 'fast_' * 51    # 255 characters
        variant = {'variant_id': variant_id}
        test_id = TestID(uid, name, variant=variant)
        self.assertEqual(test_id.uid, 1)
        self.assertEqual(test_id.str_filesystem, '%s_%s' % (uid, variant_id[:253]))
        self.assertIs(test_id.variant, variant_id)
        self.assertEqual(test_id.str_variant, ";%s" % variant_id)


if __name__ == '__main__':
    unittest.main()



# Copyright (C) IBM 2016 - Harish <harisrir@linux.vnet.ibm.com>
# Copyright (C) Red Hat 2016 - Lukas Doktor <ldoktor@redhat.com>
# Copyright (C) Intra2net AG 2018 - Plamen Dimitrov <pdimitrov@pevogam.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; version 2.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.

# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Source: https://github.com/autotest/autotest/blob/master/client/lv_utils.py


import logging
import os
import re
import shutil
import time
import warnings

from . import process

LOGGER = logging.getLogger('avocado.test')


class LVException(Exception):
    """
    Base Exception Class for all exceptions
    """


def get_diskspace(disk):
    """
    Get the entire disk space of a given disk.

    :param str disk: name of the disk to find the free space of
    :returns: size in bytes
    :rtype: str
    :raises: :py:class:`LVException` on failure to find disk space
    """
    warnings.warn("deprecated, use get_device_total_space instead",
                  DeprecationWarning)
    result = process.run('fdisk -l %s' % disk,
                         env={"LANG": "C"}, sudo=True).stdout_text
    results = result.splitlines()
    for line in results:
        if line.startswith('Disk ' + disk):
            return re.findall(r", (.*?) bytes", line)[0]
    raise LVException('Error in finding disk space')


def get_device_total_space(disk):
    """Get the total device size.

    :param str device: name of the device/disk to find the total size
    :returns: size in bytes
    :rtype: int
    :raises: :py:class:`LVException` on failure to find disk space
    """
    result = process.run('fdisk -l %s' % disk,
                         env={"LANG": "C"}, sudo=True).stdout_text
    results = result.splitlines()
    for line in results:
        if line.startswith('Disk ' + disk):
            return int(re.findall(r", (.*?) bytes", line)[0])
    raise LVException('Error in finding disk space')


def get_devices_total_space(devices):
    """Get the total size of given device(s)/disk(s).

    :param list devices: list with the names of devices separated with space.
    :returns: sizes in bytes
    :rtype: int
    :raises: :py:class:`LVException` on failure to find disk space
    """
    size = 0
    for device in devices:
        size = size + get_device_total_space(device)
    if not size:
        raise LVException('failed to get disks size')
    return size


def vg_ramdisk(disk, vg_name, ramdisk_vg_size,
               ramdisk_basedir, ramdisk_sparse_filename,
               use_tmpfs=True):
    """
    Create volume group on top of ram memory to speed up LV performance.
    When disk is specified the size of the physical volume is taken from
    existing disk space.

    :param str disk: name of the disk in which volume groups are created
    :param str vg_name: name of the volume group
    :param str ramdisk_vg_size: size of the ramdisk virtual group (MB)
    :param str ramdisk_basedir: base directory for the ramdisk sparse file
    :param str ramdisk_sparse_filename: name of the ramdisk sparse file
    :param bool use_tmpfs: whether to use RAM or slower storage
    :returns: ramdisk_filename, vg_ramdisk_dir, vg_name, loop_device
    :rtype: (str, str, str, str)
    :raises: :py:class:`LVException` on failure at any stage

    Sample ramdisk params:
    - ramdisk_vg_size = "40000"
    - ramdisk_basedir = "/tmp"
    - ramdisk_sparse_filename = "virtual_hdd"

    Sample general params:
    - vg_name='autotest_vg',
    - lv_name='autotest_lv',
    - lv_size='1G',
    - lv_snapshot_name='autotest_sn',
    - lv_snapshot_size='1G'
    The ramdisk volume group size is in MB.
    """
    warnings.warn("deprecated, use existing methods: vg_create, lv_create",
                  DeprecationWarning)
    vg_size = ramdisk_vg_size
    vg_ramdisk_dir = os.path.join(ramdisk_basedir, vg_name)
    ramdisk_filename = os.path.join(vg_ramdisk_dir,
                                    ramdisk_sparse_filename)
    # Try to cleanup the ramdisk before defining it
    try:
        vg_ramdisk_cleanup(ramdisk_filename, vg_ramdisk_dir,
                           vg_name, use_tmpfs)
    except LVException:
        pass
    if not os.path.exists(vg_ramdisk_dir):
        os.makedirs(vg_ramdisk_dir)
    try:
        if use_tmpfs:
            LOGGER.debug("Mounting tmpfs")
            process.run("mount -t tmpfs tmpfs %s" % vg_ramdisk_dir,
                        sudo=True)

        LOGGER.debug("Converting and copying /dev/zero")
        if disk:
            vg_size = get_diskspace(disk)

        # Initializing sparse file with extra few bytes
        cmd = ("dd if=/dev/zero of=%s bs=1M count=1 seek=%s" %
               (ramdisk_filename, vg_size))
        process.run(cmd)
        if not disk:
            LOGGER.debug("Finding free loop device")
            result = process.run("losetup --find", sudo=True)
    except process.CmdError as ex:
        LOGGER.error(ex)
        vg_ramdisk_cleanup(ramdisk_filename, vg_ramdisk_dir,
                           vg_name, use_tmpfs)
        raise LVException("Fail to create vg_ramdisk: %s" % ex)

    if not disk:
        loop_device = result.stdout_text.rstrip()
    else:
        loop_device = disk
    try:
        if not disk:
            LOGGER.debug("Creating loop device")
            process.run("losetup %s %s" %
                        (loop_device, ramdisk_filename), sudo=True)
        LOGGER.debug("Creating physical volume %s", loop_device)
        process.run("pvcreate -y %s" % loop_device, sudo=True)
        LOGGER.debug("Creating volume group %s", vg_name)
        process.run("vgcreate %s %s" %
                    (vg_name, loop_device), sudo=True)
    except process.CmdError as ex:
        LOGGER.error(ex)
        vg_ramdisk_cleanup(ramdisk_filename, vg_ramdisk_dir,
                           vg_name, loop_device, use_tmpfs)
        raise LVException("Fail to create vg_ramdisk: %s" % ex)
    return ramdisk_filename, vg_ramdisk_dir, vg_name, loop_device


def vg_ramdisk_cleanup(ramdisk_filename=None, vg_ramdisk_dir=None,
                       vg_name=None, loop_device=None, use_tmpfs=True):
    """
    Clean up any stage of the VG ramdisk setup in case of test error.

    This detects whether the components were initialized and if so tries
    to remove them. In case of failure it raises summary exception.

    :param str ramdisk_filename: name of the ramdisk sparse file
    :param str vg_ramdisk_dir: location of the ramdisk file
    :param str vg_name: name of the volume group
    :param str loop_device: name of the disk or loop device
    :param bool use_tmpfs: whether to use RAM or slower storage
    :returns: ramdisk_filename, vg_ramdisk_dir, vg_name, loop_device
    :rtype: (str, str, str, str)
    :raises: :py:class:`LVException` on intolerable failure at any stage
    """
    warnings.warn("deprecated, use existing methods: vg_remove, lv_remove",
                  DeprecationWarning)
    errs = []
    if vg_name is not None:
        loop_device = re.search(r"([/\w-]+) +%s +lvm2" % vg_name,
                                process.run("pvs", sudo=True).stdout_text)
        if loop_device is not None:
            loop_device = loop_device.group(1)
        process.run("vgremove -f %s" %
                    vg_name, ignore_status=True, sudo=True)

    if loop_device is not None:
        result = process.run("pvremove %s" % loop_device,
                             ignore_status=True, sudo=True)
        if result.exit_status != 0:
            errs.append("wipe pv")
            LOGGER.error("Failed to wipe pv from %s: %s", loop_device, result)

        losetup_all = process.run("losetup --all", sudo=True).stdout_text
        if loop_device in losetup_all:
            ramdisk_filename = re.search(r"%s: \[\d+\]:\d+ \(([/\w]+)\)" %
                                         loop_device, losetup_all)
            if ramdisk_filename is not None:
                ramdisk_filename = ramdisk_filename.group(1)

            for _ in range(10):
                result = process.run("losetup -d %s" % loop_device,
                                     ignore_status=True, sudo=True)
                if b"resource busy" not in result.stderr:
                    if result.exit_status != 0:
                        errs.append("remove loop device")
                        LOGGER.error("Unexpected failure when removing loop"
                                     "device %s, check the log", loop_device)
                    break
                time.sleep(0.1)

    if ramdisk_filename is not None:
        if os.path.exists(ramdisk_filename):
            os.unlink(ramdisk_filename)
            LOGGER.debug("Ramdisk filename %s deleted", ramdisk_filename)
            vg_ramdisk_dir = os.path.dirname(ramdisk_filename)

    if vg_ramdisk_dir is not None:
        if use_tmpfs and not process.system("mountpoint %s" % vg_ramdisk_dir,
                                            ignore_status=True):
            for _ in range(10):
                result = process.run("umount %s" % vg_ramdisk_dir,
                                     ignore_status=True, sudo=True)
                time.sleep(0.1)
                if result.exit_status == 0:
                    break
            else:
                errs.append("umount")
                LOGGER.error("Unexpected failure unmounting %s, check the "
                             "log", vg_ramdisk_dir)

        if os.path.exists(vg_ramdisk_dir):
            try:
                shutil.rmtree(vg_ramdisk_dir)
                LOGGER.debug("Ramdisk directory %s deleted", vg_ramdisk_dir)
            except OSError as details:
                errs.append("rm-ramdisk-dir")
                LOGGER.error("Failed to remove ramdisk_dir: %s", details)
    if errs:
        raise LVException("vg_ramdisk_cleanup failed: %s" % ", ".join(errs))


def vg_check(vg_name):
    """
    Check whether provided volume group exists.

    :param str vg_name: name of the volume group
    :returns: whether the volume group was found
    :rtype: bool
    """
    cmd = "vgdisplay %s" % vg_name
    try:
        process.run(cmd, sudo=True)
        LOGGER.debug("Provided volume group exists: %s", vg_name)
        return True
    except process.CmdError as exception:
        LOGGER.error(exception)
        return False


def vg_list(vg_name=None):
    """
    List all info about available volume groups.

    :param vg_name: name of the volume group to list or or None to list all
    :type vg_name: str or None
    :returns: list of available volume groups
    :rtype: {str, {str, str}}
    """
    cmd = "vgs --all"
    cmd += " %s" % vg_name if vg_name is not None else ""
    vgroups = {}
    result = process.run(cmd, sudo=True)
    lines = result.stdout_text.strip().splitlines()
    if len(lines) > 1:
        columns = lines[0].split()
        lines = lines[1:]
    else:
        return vgroups
    # TODO: Optimize this
    for line in lines:
        details = line.split()
        details_dict = {}
        index = 0
        for column in columns:
            if "VG" in column:
                vg_name = details[index]
            else:
                details_dict[column] = details[index]
            index += 1
        vgroups[vg_name] = details_dict
    return vgroups


def vg_create(vg_name, pv_list, force=False):
    """
    Create a volume group from a list of physical volumes.

    :param str vg_name: name of the volume group
    :param pv_list: list of physical volumes to use
    :type pv_list: str or [str]
    :param bool force: create volume group with a force flag
    :raises: :py:class:`LVException` if volume group already exists
    """
    if vg_check(vg_name):
        raise LVException("Volume group '%s' already exist" % vg_name)
    cmd = "vgcreate"
    if isinstance(pv_list, list):
        pv_list = " ".join(pv_list)
    else:
        pv_list = str(pv_list)
    cmd += " %s %s" % (vg_name, pv_list)
    if force:
        cmd += "%s -f -y" % cmd
    process.run(cmd, sudo=True)


def vg_remove(vg_name):
    """
    Remove a volume group.

    :param str vg_name: name of the volume group
    :raises: :py:class:`LVException` if volume group cannot be found
    """
    if not vg_check(vg_name):
        raise LVException("Volume group '%s' could not be found" % vg_name)
    cmd = "vgremove -f %s" % vg_name
    process.run(cmd, sudo=True)


def lv_check(vg_name, lv_name):
    """
    Check whether provided logical volume exists.

    :param str vg_name: name of the volume group
    :param str lv_name: name of the logical volume
    :return: whether the logical volume was found
    :rtype: bool
    """
    cmd = "lvdisplay %s" % vg_name
    result = process.run(cmd, ignore_status=True, sudo=True)

    lvpattern = r"LV Name\s+%s\s+" % lv_name
    match = re.search(lvpattern, result.stdout_text.rstrip())
    if match:
        LOGGER.debug("Provided Logical volume %s exists in %s",
                     lv_name, vg_name)
        return True
    else:
        return False


def lv_list(vg_name=None):
    """
    List all info about available logical volumes.

    :param str vg_name: name of the volume group or None to list all
    :returns: list of available logical volumes
    :rtype: {str, {str, str}}
    """
    cmd = "lvs --all"
    cmd += " %s" % vg_name if vg_name is not None else ""
    volumes = {}
    result = process.run(cmd, sudo=True)

    lines = result.stdout_text.strip().splitlines()
    if len(lines) > 1:
        lines = lines[1:]
    else:
        return volumes

    for line in lines:
        details = line.split()
        length = len(details)
        details_dict = {}
        lv_name = details[0]
        details_dict["VG"] = details[1]
        details_dict["Attr"] = details[2]
        details_dict["LSize"] = details[3]
        if length == 5:
            details_dict["Origin_Data"] = details[4]
        elif length > 5:
            details_dict["Origin_Data"] = details[5]
            details_dict["Pool"] = details[4]
        volumes[lv_name] = details_dict
    return volumes


def lv_create(vg_name, lv_name, lv_size, force_flag=True,
              pool_name=None, pool_size="1G"):
    """
    Create a (possibly thin) logical volume in a volume group.
    The volume group must already exist.

    A thin pool will be created if pool parameters are provided
    and the thin pool doesn't already exist.

    The volume group must already exist.

    :param str vg_name: name of the volume group
    :param str lv_name: name of the logical volume
    :param str lv_size: size for the logical volume to be created
    :param bool force_flag: whether to abort if volume already exists
                            or remove and recreate it
    :param str pool_name: name of thin pool or None for a regular volume
    :param str pool_size: size of thin pool if it will be created
    :raises: :py:class:`LVException` if preconditions or execution fails
    """
    if not vg_check(vg_name):
        raise LVException("Volume group could not be found")
    if lv_check(vg_name, lv_name) and not force_flag:
        raise LVException("Logical volume already exists")
    elif lv_check(vg_name, lv_name) and force_flag:
        lv_remove(vg_name, lv_name)

    lv_cmd = "lvcreate --name %s" % lv_name
    if pool_name is not None:
        if not lv_check(vg_name, pool_name):
            tp_cmd = "lvcreate --thinpool %s --size %s %s -y" % (pool_name,
                                                                 pool_size,
                                                                 vg_name)
            try:
                process.run(tp_cmd, sudo=True)
            except process.CmdError as detail:
                LOGGER.debug(detail)
                raise LVException("Create thin volume pool failed.")
            LOGGER.debug("Created thin volume pool: %s", pool_name)
        lv_cmd += " --virtualsize %s" % lv_size
        lv_cmd += " --thin %s/%s -y" % (vg_name, pool_name)
    else:
        lv_cmd += " --size %s" % lv_size
        lv_cmd += " %s -y" % vg_name
    try:
        process.run(lv_cmd, sudo=True)
    except process.CmdError as detail:
        LOGGER.error(detail)
        raise LVException("Create thin volume failed.")
    LOGGER.debug("Created thin volume:%s", lv_name)


def lv_remove(vg_name, lv_name):
    """
    Remove a logical volume.

    :param str vg_name: name of the volume group
    :param str lv_name: name of the logical volume
    :raises: :py:class:`LVException` if volume group or logical
             volume cannot be found
    """
    if not vg_check(vg_name):
        raise LVException("Volume group could not be found")
    if not lv_check(vg_name, lv_name):
        raise LVException("Logical volume could not be found")

    cmd = "lvremove -f %s/%s" % (vg_name, lv_name)
    process.run(cmd, sudo=True)


def lv_take_snapshot(vg_name, lv_name,
                     lv_snapshot_name, lv_snapshot_size=None,
                     pool_name=None):
    """
    Take a (possibly thin) snapshot of a regular (or thin) logical volume.

    :param str vg_name: name of the volume group
    :param str lv_name: name of the logical volume
    :param str lv_snapshot_name: name of the snapshot be to created
    :param str lv_snapshot_size: size of the snapshot or None for thin
                                 snapshot of an already thin volume
    :param pool_name: name of thin pool or None for regular snapshot
                      or snapshot in the same thin pool like the volume
    :raises: :py:class:`process.CmdError` on failure to create snapshot
    :raises: :py:class:`LVException` if preconditions fail
    """
    if not vg_check(vg_name):
        raise LVException("Volume group could not be found")
    if pool_name is not None and not lv_check(vg_name, pool_name):
        raise LVException("Snapshot's thin pool could not be found")
    if lv_check(vg_name, lv_snapshot_name):
        raise LVException("Snapshot already exists")
    if not lv_check(vg_name, lv_name):
        raise LVException("Snapshot's origin could not be found")

    # thin snapshot extensions (from thin or external volume)
    cmd = ("lvcreate --snapshot --name %s /dev/%s/%s --ignoreactivationskip" %
           (lv_snapshot_name, vg_name, lv_name))
    if lv_snapshot_size is not None:
        cmd += " --size %s" % lv_snapshot_size
    if pool_name is not None:
        cmd += " --thinpool %s/%s" % (vg_name, pool_name)

    try:
        process.run(cmd, sudo=True)
    except process.CmdError as ex:
        lv = 'Logical volume "%s" already exists in volume group "%s"' % (lv_snapshot_name, vg_name)
        if lv in ex.result.stderr_text:
            active = lv_snapshot_name + " [active]" in process.run("lvdisplay", sudo=True).stdout_text
            if active:
                # the above conditions detect if merge of snapshot was postponed
                log_msg = "Logical volume %s is still active! Attempting to deactivate..."
                LOGGER.debug(log_msg, lv_name)
                lv_reactivate(vg_name, lv_name)
                process.run(cmd, sudo=True)
        else:
            raise ex


def lv_revert(vg_name, lv_name, lv_snapshot_name):
    """
    Revert the origin logical volume to a snapshot.

    :param str vg_name: name of the volume group
    :param str lv_name: name of the logical volume
    :param str lv_snapshot_name: name of the snapshot to be reverted
    :raises: :py:class:`process.CmdError` on failure to revert snapshot
    :raises: :py:class:`LVException` if preconditions or execution fails
    """
    try:
        if not vg_check(vg_name):
            raise LVException("Volume group could not be found")
        if not lv_check(vg_name, lv_snapshot_name):
            raise LVException("Snapshot could not be found")
        if (not lv_check(vg_name, lv_snapshot_name) and not lv_check(vg_name,
                                                                     lv_name)):
            raise LVException("Snapshot and its origin could not be found")
        if (lv_check(vg_name, lv_snapshot_name) and not lv_check(vg_name,
                                                                 lv_name)):
            raise LVException("Snapshot origin could not be found")

        cmd = ("lvconvert --merge --interval 1 /dev/%s/%s" % (vg_name, lv_snapshot_name))
        result = process.run(cmd, sudo=True)
        if (("Merging of snapshot %s will start next activation." %
             lv_snapshot_name) in result.stdout_text):
            raise LVException("The Logical volume %s is still active" %
                              lv_name)

    except process.CmdError as ex:
        # detect if merge of snapshot was postponed
        # and attempt to reactivate the volume.
        active_lv_pattern = re.escape("%s [active]" % lv_snapshot_name)
        lvdisplay_output = process.run("lvdisplay", sudo=True).stdout_text
        if ('Snapshot could not be found' in ex.result.stderr_text and
                re.search(active_lv_pattern, lvdisplay_output) or
                "The Logical volume %s is still active" % lv_name in ex.result.stderr_text):
            log_msg = "Logical volume %s is still active! Attempting to deactivate..."
            LOGGER.debug(log_msg, lv_name)
            lv_reactivate(vg_name, lv_name)
            LOGGER.error("Continuing after reactivation")
        elif 'Snapshot could not be found' in ex.result.stderr_text:
            LOGGER.error("Could not revert to snapshot:")
            LOGGER.error(ex.result)
        else:
            raise ex


def lv_revert_with_snapshot(vg_name, lv_name,
                            lv_snapshot_name, lv_snapshot_size):
    """
    Perform logical volume merge with snapshot and take a new snapshot.

    :param str vg_name: name of the volume group
    :param str lv_name: name of the logical volume
    :param str lv_snapshot_name: name of the snapshot to be reverted
    :param str lv_snapshot_size: size of the snapshot
    """
    lv_revert(vg_name, lv_name, lv_snapshot_name)
    lv_take_snapshot(vg_name, lv_name, lv_snapshot_name, lv_snapshot_size)


def lv_reactivate(vg_name, lv_name, timeout=10):
    """
    In case of unclean shutdowns some of the lvs is still active and merging
    is postponed. Use this function to attempt to deactivate and reactivate
    all of them to cause the merge to happen.

    :param str vg_name: name of the volume group
    :param str lv_name: name of the logical volume
    :param int timeout: timeout between operations
    :raises: :py:class:`LVException` if the logical volume is still active
    """
    try:
        process.run("lvchange -an /dev/%s/%s" % (vg_name, lv_name), sudo=True)
        time.sleep(timeout)
        process.run("lvchange -ay /dev/%s/%s" % (vg_name, lv_name), sudo=True)
        time.sleep(timeout)
    except process.CmdError:
        log_msg = "Failed to reactivate %s - please, nuke the process that uses it first."
        LOGGER.error(log_msg, lv_name)
        raise LVException("The Logical volume %s is still active" % lv_name)


def lv_mount(vg_name, lv_name, mount_loc, create_filesystem=""):
    """
    Mount a logical volume to a mount location.

    :param str vg_name: name of the volume group
    :param str lv_name: name of the logical volume
    :param str mount_loc: location to mount the logical volume to
    :param str create_filesystem: can be one of ext2, ext3, ext4, vfat or empty
                                  if the filesystem was already created and the
                                  mkfs process is skipped
    :raises: :py:class:`LVException` if the logical volume could not be mounted
    """
    try:
        if create_filesystem:
            process.run("mkfs.%s /dev/%s/%s" %
                        (create_filesystem, vg_name, lv_name),
                        sudo=True)
        process.run("mount /dev/%s/%s %s" %
                    (vg_name, lv_name, mount_loc), sudo=True)
    except process.CmdError as ex:
        raise LVException("Fail to mount logical volume: %s" % ex)


def vg_reactivate(vg_name, timeout=10, export=False):
    """
    In case of unclean shutdowns some of the vgs is still active and merging
    is postponed. Use this function to attempt to deactivate and reactivate
    all of them to cause the merge to happen.

    :param str vg_name: name of the volume group
    :param int timeout: timeout between operations
    :raises: :py:class:`LVException` if the logical volume is still active
    """
    try:
        process.run("vgchange -an %s" % vg_name, sudo=True)
        time.sleep(timeout)

        if export:
            process.run("vgexport %s" % vg_name, sudo=True)
            time.sleep(timeout)
            process.run("vgimport %s" % vg_name, sudo=True)
            time.sleep(timeout)

        process.run("vgchange -ay %s" % vg_name, sudo=True)
        time.sleep(timeout)
    except process.CmdError:
        log_msg = "Failed to reactivate %s - please, nuke the process that uses it first."
        LOGGER.error(log_msg, vg_name)
        raise LVException("The Volume group %s is still active" % vg_name)


def lv_umount(vg_name, lv_name):
    """
    Unmount a Logical volume from a mount location.

    :param str vg_name: name of the volume group
    :param str lv_name: name of the logical volume
    :raises: :py:class:`LVException` if the logical volume could not be unmounted
    """
    try:
        process.run("umount /dev/%s/%s" % (vg_name, lv_name), sudo=True)
    except process.CmdError as ex:
        raise LVException("Fail to unmount logical volume: %s" % ex)



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat Inc. 2020
# Authors: Cleber Rosa <crosa@redhat.com>

from avocado.core import varianter
from avocado.core.plugin_interfaces import Init, Varianter
from avocado.core.settings import settings
from avocado.core.tree import TreeNode


class DictVariantsInit(Init):

    name = 'dict_variants'
    description = "Python Dictionary based varianter"

    def initialize(self):
        help_msg = 'Load the Variants from Python dictionaries'
        settings.register_option(section='run',
                                 key='dict_variants',
                                 default=[],
                                 key_type=list,
                                 help_msg=help_msg)


class DictVariants(Varianter):

    """
    Turns (a list of) Python dictionaries into variants
    """

    name = 'dict_variants'
    description = "Python Dictionary based varianter"

    def initialize(self, config):
        # pylint: disable=W0201
        self.variants = config.get('run.dict_variants')
        if self.variants:
            # pylint: disable=W0201
            self.headers = [key for key in self.variants[0].keys()]

    def __iter__(self):
        if self.variants is None:
            return

        variant_ids = []
        for variant in self.variants:
            variant_ids.append("-".join([str(variant.get(key))
                                         for key in self.headers]))

        for vid, variant in zip(variant_ids, self.variants):
            yield {"variant_id": vid,
                   "variant": [TreeNode('', variant)],
                   "paths": ['/']}

    def __len__(self):
        return sum(1 for _ in self.variants) if self.variants else 0

    def to_str(self, summary, variants, **kwargs):
        """
        Return human readable representation

        The summary/variants accepts verbosity where 0 means silent and
        maximum is up to the plugin.

        :param summary: How verbose summary to output (int)
        :param variants: How verbose list of variants to output (int)
        :param kwargs: Other free-form arguments
        :rtype: str
        """
        if not self.variants:
            return ""
        out = []

        if variants:
            # variants == 0 means disable, but in plugin it's brief
            out.append("Dict Variants (%s):" % len(self))
            for variant in self:
                out.extend(varianter.variant_to_str(variant, variants - 1,
                                                    kwargs, False))
        return "\n".join(out)



#!/usr/bin/env python3

import sys

from avocado.core.job import Job
from avocado.core.suite import TestSuite

config = {'run.references': ['examples/tests/sleeptest.py:SleepTest.test'],
          'json.variants.load': 'examples/tests/sleeptest.py.data/sleeptest.json'}

suite = TestSuite.from_config(config)
with Job(config, [suite]) as j:
    sys.exit(j.run())



#!/bin/env python3

import os
import sys

from avocado.core.job import Job

THIS_DIR = os.path.dirname(os.path.abspath(__file__))
ROOT_DIR = os.path.dirname(os.path.dirname(os.path.dirname(THIS_DIR)))
TESTS_DIR = os.path.join(os.path.dirname(THIS_DIR), 'tests')

cirrus_ci = {
    'run.references': [os.path.join(TESTS_DIR, 'cirrusci.py')],
    }

parallel_1 = {
    'run.test_runner': 'nrunner',
    'run.references': [os.path.join('selftests', 'unit'),
                       os.path.join('selftests', 'functional')],
    'filter.by_tags.tags': ['parallel:1'],
    'nrunner.max_parallel_tasks': 1,
    }

vmimage = {
    'run.references': [os.path.join(TESTS_DIR, 'vmimage.py')],
    'yaml_to_mux.files': [os.path.join(TESTS_DIR, 'vmimage.py.data', 'variants.yml')],
    }

if __name__ == '__main__':
    os.chdir(ROOT_DIR)
    with Job.from_config({}, [cirrus_ci, parallel_1, vmimage]) as j:
        os.environ['AVOCADO_CHECK_LEVEL'] = '3'
        sys.exit(j.run())



from avocado import Test


class CitParameters(Test):

    """
    Example test that fetches the parameters set on examples/cit/params.ini

    :avocado: tags=fast
    """

    def test(self):
        self.params.get('color')
        self.params.get('shape')
        self.params.get('state')
        self.params.get('material')
        self.params.get('coating')



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat Inc. 2014
#
# Authors: Ruda Moura <rmoura@redhat.com>
#          Lucas Meneghel Rodrigues <lmr@redhat.com>
#          Harish S <harisrir@linux.vnet.ibm.com>
#

"""
This module contains handy classes that can be used inside
avocado core code or plugins.
"""


import math
import sys


class InvalidDataSize(ValueError):
    """
    Signals that the value given to :class:`DataSize` is not valid.
    """


def ordered_list_unique(object_list):
    """
    Returns an unique list of objects, with their original order preserved
    """
    seen = set()
    seen_add = seen.add
    return [x for x in object_list if not (x in seen or seen_add(x))]


def geometric_mean(values):
    """
    Evaluates the geometric mean for a list of numeric values.
    This implementation is slower but allows unlimited number of values.
    :param values: List with values.
    :return: Single value representing the geometric mean for the list values.
    :see: http://en.wikipedia.org/wiki/Geometric_mean
    """
    try:
        values = [int(value) for value in values]
    except ValueError:
        raise ValueError('Invalid inputs %s. Provide valid inputs' % values)
    no_values = len(values)
    if no_values == 0:
        return None
    return math.exp(sum([math.log(number) for number in values]) / no_values)


def compare_matrices(matrix1, matrix2, threshold=0.05):
    """
    Compare 2 matrices nxm and return a matrix nxm with comparison data and
    stats. When the first columns match, they are considered as header and
    included in the results intact.

    :param matrix1: Reference Matrix of floats; first column could be header.
    :param matrix2: Matrix that will be compared; first column could be header
    :param threshold: Any difference greater than this percent threshold will
                      be reported.
    :return: Matrix with the difference in comparison, number of improvements,
             number of regressions, total number of comparisons.
    """
    improvements = 0
    regressions = 0
    same = 0
    new_matrix = []

    for line1, line2 in zip(matrix1, matrix2):
        new_line = []
        elements = iter(zip(line1, line2))
        try:
            element1, element2 = next(elements)
        except StopIteration:             # no data in this row
            new_matrix.append(new_line)
            continue
        if element1 == element2:          # this column contains header
            new_line.append(element1)
            try:
                element1, element2 = next(elements)
            except StopIteration:
                new_matrix.append(new_line)
                continue
        while True:
            try:
                ratio = float(element2) / float(element1)
            except ZeroDivisionError:  # For 0s, allow exact match or error
                if float(element2) == 0:
                    new_line.append(".")
                    same += 1
                else:
                    new_line.append("error_%s/%s" % (element2, element1))
                    improvements += 1
                try:
                    element1, element2 = next(elements)
                except StopIteration:
                    break
                continue
            if ratio < (1 - threshold):     # handling regression
                regressions += 1
                new_line.append(100 * ratio - 100)
            elif ratio > (1 + threshold):   # handling improvements
                improvements += 1
                new_line.append("+%.6g" % (100 * ratio - 100))
            else:
                same += 1
                new_line.append(".")
            try:
                element1, element2 = next(elements)
            except StopIteration:
                break
        new_matrix.append(new_line)

    total = improvements + regressions + same
    return (new_matrix, improvements, regressions, total)


def comma_separated_ranges_to_list(string):
    """
    Provides a list from comma separated ranges

    :param string: string of comma separated range
    :return list: list of integer values in comma separated range
    """
    values = []
    for value in string.split(','):
        if '-' in value:
            start, end = value.split('-')
            for val in range(int(start), int(end) + 1):
                values.append(int(val))
        else:
            values.append(int(value))
    return values


class Borg:

    """
    Multiple instances of this class will share the same state.

    This is considered a better design pattern in Python than
    more popular patterns, such as the Singleton. Inspired by
    Alex Martelli's article mentioned below:

    :see: http://www.aleax.it/5ep.html
    """

    __shared_state = {}

    def __init__(self):
        self.__dict__ = self.__shared_state


class LazyProperty:

    """
    Lazily instantiated property.

    Use this decorator when you want to set a property that will only be
    evaluated the first time it's accessed. Inspired by the discussion in
    the Stack Overflow thread below:

    :see: http://stackoverflow.com/questions/15226721/
    """

    def __init__(self, f_get):
        self.f_get = f_get
        self.func_name = f_get.__name__

    def __get__(self, obj, cls):
        if obj is None:
            return None
        value = self.f_get(obj)
        setattr(obj, self.func_name, value)
        return value


class CallbackRegister:

    """
    Registers pickable functions to be executed later.
    """

    def __init__(self, name, log):
        """
        :param name: Human readable identifier of this register
        """
        self._name = name
        self._items = []
        self._log = log

    def register(self, func, args, kwargs, once=False):
        """
        Register function/args to be called on self.destroy()
        :param func: Pickable function
        :param args: Pickable positional arguments
        :param kwargs: Pickable keyword arguments
        :param once: Add unique (func,args,kwargs) combination only once
        """
        item = (func, args, kwargs)
        if not once or item not in self._items:
            self._items.append(item)

    def unregister(self, func, args, kwargs):
        """
        Unregister (func,args,kwargs) combination
        :param func: Pickable function
        :param args: Pickable positional arguments
        :param kwargs: Pickable keyword arguments
        """
        item = (func, args, kwargs)
        if item in self._items:
            self._items.remove(item)

    def run(self):
        """
        Call all registered function
        """
        while self._items:
            item = self._items.pop()
            try:
                func, args, kwargs = item
                func(*args, **kwargs)
            except:     # Ignore all exceptions pylint: disable=W0702
                self._log.error("%s failed to destroy %s:\n%s",
                                self._name, item, sys.exc_info()[1])

    def __del__(self):
        """
        :warning: Always call self.run() manually, this is not guaranteed
                  to be executed!
        """
        self.run()


def time_to_seconds(time):
    """
    Convert time in minutes, hours and days to seconds.
    :param time: Time, optionally including the unit (i.e. '10d')
    """
    units = {'s': 1, 'm': 60, 'h': 3600, 'd': 86400}
    if time is not None:
        try:
            unit = time[-1].lower()
            if unit in units:
                mult = units[unit]
                seconds = int(time[:-1]) * mult
            else:
                seconds = int(time)
        except (ValueError, TypeError):
            raise ValueError("Invalid value '%s' for time. Use a string with "
                             "the number and optionally the time unit (s, m, "
                             "h or d)." % time)
    else:
        seconds = 0
    return seconds


class DataSize:
    """
    Data Size object with builtin unit-converted attributes.

    :param data: Data size plus optional unit string. i.e. '10m'. No
                 unit string means the data size is in bytes.
    :type data: str
    """

    __slots__ = ['_value', '_unit']

    MULTIPLIERS = {'b': 1,  # 2**0
                   'k': 1024,  # 2**10
                   'm': 1048576,  # 2**20
                   'g': 1073741824,  # 2**30
                   't': 1099511627776}  # 2**40

    def __init__(self, data):
        try:
            norm_size = data.strip().lower()
            last = norm_size[-1]
            if last.isdigit():
                self._value = int(norm_size)
                self._unit = 'b'
            elif last in self.MULTIPLIERS:
                self._value = int(norm_size[:-1])
                self._unit = last
            else:
                raise ValueError

            if self._value < 0:
                raise ValueError

        except ValueError:
            raise InvalidDataSize('String not in size + unit format (i.e. '
                                  '"10M", "100k", ...)')

    @property
    def value(self):
        return self._value

    @property
    def unit(self):
        return self._unit

    @property
    def b(self):
        return self._value * self.MULTIPLIERS[self._unit]

    @property
    def k(self):
        return int(self._value * self.MULTIPLIERS[self._unit] /
                   self.MULTIPLIERS['k'])

    @property
    def m(self):
        return int(self._value * self.MULTIPLIERS[self._unit] /
                   self.MULTIPLIERS['m'])

    @property
    def g(self):
        return int(self._value * self.MULTIPLIERS[self._unit] /
                   self.MULTIPLIERS['g'])

    @property
    def t(self):
        return int(self._value * self.MULTIPLIERS[self._unit] /
                   self.MULTIPLIERS['t'])



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat Inc. 2015
# Author: Cleber Rosa <cleber@redhat.com>


import abc


class Plugin(metaclass=abc.ABCMeta):
    """Base for all plugins."""


class Init(Plugin):
    """Base plugin interface for plugins that needs to initialize itself."""

    @abc.abstractmethod
    def initialize(self):
        """Entry point for the plugin to perform its initialization."""


class Settings(Plugin):
    """Base plugin to allow modifying settings.

    Currently it only supports to extend/modify the default list of
    paths to config files.
    """

    @abc.abstractmethod
    def adjust_settings_paths(self, paths):
        """Entry point where plugin can modify the list of configuration paths."""


class CLI(Plugin):
    """Base plugin interface for adding options (non-commands) to the command line.

    Plugins that want to add extra options to the core command line application
    or to sub commands should use the 'avocado.plugins.cli' namespace.
    """

    @abc.abstractmethod
    def configure(self, parser):
        """Configures the command line parser with options specific to this plugin."""

    @abc.abstractmethod
    def run(self, config):
        """Execute any action the plugin intends.

        Example of action may include activating a special features upon
        finding that the requested command line options were set by the user.

        Note: this plugin class is not intended for adding new commands, for
        that please use `CLICmd`.
        """


class CLICmd(Plugin):
    """Base plugin interface for adding new commands to the command line app.

    Plugins that want to add extensions to the run command should use the
    'avocado.plugins.cli.cmd' namespace.
    """
    name = None
    description = None

    def configure(self, parser):
        """Lets the extension add command line options and do early configuration.

        By default it will register its `name` as the command name and give
        its `description` as the help message.
        """
        help_msg = self.description
        if help_msg is None:
            help_msg = 'Runs the %s command' % self.name

        parser = parser.subcommands.add_parser(self.name,
                                               help=help_msg)
        return parser

    @abc.abstractmethod
    def run(self, config):
        """Entry point for actually running the command."""


class JobPre(Plugin):
    """Base plugin interface for adding actions before a job runs.

    Plugins that want to add actions to be run before a job runs,
    should use the 'avocado.plugins.job.prepost' namespace and
    implement the defined interface.
    """

    @abc.abstractmethod
    def pre(self, job):
        """Entry point for actually running the pre job action."""


class JobPost(Plugin):
    """Base plugin interface for adding actions after a job runs.

    Plugins that want to add actions to be run after a job runs,
    should use the 'avocado.plugins.job.prepost' namespace and
    implement the defined interface.
    """

    @abc.abstractmethod
    def post(self, job):
        """Entry point for actually running the post job action."""


class Result(Plugin):

    @abc.abstractmethod
    def render(self, result, job):
        """Entry point with method that renders the result.

        This will usually be used to write the result to a file or directory.

        :param result: the complete job result
        :type result: :class:`avocado.core.result.Result`
        :param job: the finished job for which a result will be written
        :type job: :class:`avocado.core.job.Job`
        """


class JobPreTests(Plugin):
    """Base plugin interface for adding actions before a job runs tests.

    This interface looks similar to :class:`JobPre`, but it's intended
    to be called at a very specific place, that is, between
    :meth:`avocado.core.job.Job.create_test_suite` and
    :meth:`avocado.core.job.Job.run_tests`.
    """

    @abc.abstractmethod
    def pre_tests(self, job):
        """Entry point for job running actions before tests execution."""


class JobPostTests(Plugin):
    """Base plugin interface for adding actions after a job runs tests.

    Plugins using this interface will run at the a time equivalent to
    plugins using the :class:`JobPost` interface, that is, at
    :meth:`avocado.core.job.Job.post_tests`.  This is because
    :class:`JobPost` based plugins will eventually be modified to
    really run after the job has finished, and not after it has run
    tests.
    """

    @abc.abstractmethod
    def post_tests(self, job):
        """Entry point for job running actions after the tests execution."""


class ResultEvents(JobPreTests, JobPostTests):
    """Base plugin interface for event based (stream-able) results.

    Plugins that want to add actions to be run after a job runs,
    should use the 'avocado.plugins.result_events' namespace and
    implement the defined interface.
    """

    @abc.abstractmethod
    def start_test(self, result, state):
        """Event triggered when a test starts running."""

    @abc.abstractmethod
    def test_progress(self, progress=False):
        """Interface to notify progress (or not) of the running test."""

    @abc.abstractmethod
    def end_test(self, result, state):
        """Event triggered when a test finishes running."""


class Varianter(Plugin):
    """Base plugin interface for producing test variants."""

    @abc.abstractmethod
    def __iter__(self):
        """Yields all variants.

        The variant is defined as dictionary with at least:
         * variant_id - name of the current variant
         * variant - AvocadoParams-compatible variant (usually a list)
         * paths - default path(s)

        :yield variant
        """

    @abc.abstractmethod
    def __len__(self):
        """Report number of variants."""

    @abc.abstractmethod
    def to_str(self, summary, variants, **kwargs):
        """Return human readable representation.

        The summary/variants accepts verbosity where 0 means silent and
        maximum is up to the plugin.

        :param summary: How verbose summary to output (int)
        :param variants: How verbose list of variants to output (int)
        :param kwargs: Other free-form arguments
        :rtype: str
        """


class Resolver(Plugin):
    """Base plugin interface for resolving test references into resolutions."""

    @abc.abstractmethod
    def resolve(self, reference):
        """Resolves the given reference into a reference resolution.

        :param reference: a specification that can eventually be resolved
                          into a test (in the form of a
                          :class:`avocado.core.nrunner.Runnable`)
        :type reference: str
        :returns: the result of the resolution process, containing the
                  success, failure or error, along with zero or more
                  :class:`avocado.core.nrunner.Runnable` objects
        :rtype: :class:`avocado.core.resolver.ReferenceResolution`
        """


class Runner(Plugin):
    """Base plugin interface for test runners.

    This is the interface a job uses to drive the tests execution via
    compliant test runners.

    NOTE: This interface is not to be confused with the internal
    interface or idiosyncrasies of the :ref:`nrunner`.
    """

    @abc.abstractmethod
    def run_suite(self, job, test_suite):
        """Run one or more tests and report with test result.

        :param job: an instance of :class:`avocado.core.job.Job`.
        :param test_suite: an instance of TestSuite with some tests to run.
        :return: a set with types of test failures.
        """


class Spawner(Plugin):
    """Base plugin interface spawners of tasks.

    A spawner implementation will spawn a runner in its intended
    location, and isolation model.  It's supposed to be generic enough
    that it can perform that in the local machine using a process as an
    isolation model, or in a virtual machine, using the virtual
    machine itself as the isolation model.
    """

    @staticmethod
    @abc.abstractmethod
    def is_task_alive(runtime_task):
        """Determines if a task is alive or not.

        :param runtime_task: wrapper for a Task with additional runtime information
        :type runtime_task: :class:`avocado.core.task.runtime.RuntimeTask`
        """

    @abc.abstractmethod
    async def spawn_task(self, runtime_task):
        """Spawns a task return whether the spawning was successful.

        :param runtime_task: wrapper for a Task with additional runtime information
        :type runtime_task: :class:`avocado.core.task.runtime.RuntimeTask`
        """

    @abc.abstractmethod
    async def wait_task(self, runtime_task):
        """Waits for a task to finish.

        :param runtime_task: wrapper for a Task with additional runtime information
        :type runtime_task: :class:`avocado.core.task.runtime.RuntimeTask`
        """

    @staticmethod
    @abc.abstractmethod
    async def check_task_requirements(runtime_task):
        """Checks if the requirements described within a task are available.

        :param runtime_task: wrapper for a Task with additional runtime information
        :type runtime_task: :class:`avocado.core.task.runtime.RuntimeTask`
        """



import json
import re
import unittest
from urllib.error import URLError

from avocado.utils import astring, download


def get_content_by_encoding(url):
    """
    Returns the content of the given URL, attempting to use server provided
    encoding.

    :param url: the url to be fetched
    :rtype: str
    :raises: URLError when the given url can not be retrieved
    """
    http_response = download.url_open(url)
    content_type = None
    encoding = None
    if hasattr(http_response, 'headers'):
        content_type = http_response.headers['Content-Type']
    elif hasattr(http_response, 'getheader'):
        content_type = http_response.getheader('Content-Type')
    if content_type is not None:
        match = re.match(r'^[az\\].*\; charset\=(.*)$', content_type)
        if match is not None:
            encoding = match.group(1)
    content = http_response.read()
    return astring.to_text(content, encoding)


class TestThirdPartyBugs(unittest.TestCase):
    """
    Class created to verify third-party known issues
    """

    def test_inspektor_indent_bug(self):
        # https://github.com/avocado-framework/inspektor/issues/31
        # Inspektor indent will poke inside a Python string and change its
        # content.  This happened while writing selftests/unit/test_utils_cpu.py
        # with content from /proc/cpuinfo.  Right now the indent check is disabled
        # on that file
        try:
            issue_url = 'https://api.github.com/repos/avocado-framework/inspektor/issues/31'
            content = get_content_by_encoding(issue_url)
        except URLError as details:
            raise unittest.SkipTest(details)
        issue = json.loads(content)
        self.assertEqual(issue['state'], 'open', 'The issue %s is not open '
                         'anymore. Please double check and, if already fixed, '
                         'remove the selftests/unit/test_utils_cpu.py from '
                         'the exclusion list in selftests/inspekt-indent.sh ' %
                         'https://github.com/avocado-framework/inspektor/issues/31')


if __name__ == '__main__':
    unittest.main()



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat Inc. 2017
# Author: Amador Pahim <apahim@redhat.com>

"""
Provides VM images acquired from official repositories
"""

import os
import re
import tempfile
import uuid
from html.parser import HTMLParser
from urllib.error import HTTPError
from urllib.request import urlopen

from . import archive, asset, astring
from . import path as utils_path
from . import process

# pylint: disable=C0401
#: The "qemu-img" binary used when creating the snapshot images.  If
#: set to None (the default), it will attempt to find a suitable binary
#: with :func:`avocado.utils.path.find_command`, which uses the the
#: system's PATH environment variable
QEMU_IMG = None


#: Default image architecture, which defaults to the current machine
#: architecture, if that's can be retrieved via :func:`os.uname`. If
#: not, it defaults to "x86_64" simply because it's the most commonly
#: used architecture and lack of a better default
if hasattr(os, 'uname'):
    DEFAULT_ARCH = os.uname()[4]
else:
    DEFAULT_ARCH = 'x86_64'


class ImageProviderError(Exception):
    """
    Generic error class for ImageProvider
    """


class VMImageHtmlParser(HTMLParser):  # pylint: disable=W0223
    """
    Custom HTML parser to extract the href items that match
    a given pattern
    """

    def __init__(self, pattern):
        HTMLParser.__init__(self)
        self.items = []
        self.pattern = pattern

    def handle_starttag(self, tag, attrs):
        if tag != 'a':
            return
        for attr in attrs:
            if attr[0] == 'href' and re.match(self.pattern, attr[1]):
                match = attr[1].strip('/')
                if match not in self.items:
                    self.items.append(match)


class ImageProviderBase:
    """
    Base class to define the common methods and attributes of an
    image. Intended to be sub-classed by the specific image providers.
    """

    HTML_ENCODING = 'utf-8'

    def __init__(self, version, build, arch):

        self.url_versions = None
        self.url_images = None
        self.image_pattern = None
        self._file_name = None
        self._version = version
        self._best_version = None
        self.build = build
        self.arch = arch

    @property
    def version(self):
        return self._best_version or self.get_version()

    @property
    def version_pattern(self):
        return '^%s/$' % self._version

    @property
    def file_name(self):
        if self._file_name is None:
            self._file_name = self.image_pattern.format(version=self.version,
                                                        build=self.build,
                                                        arch=self.arch)
        return self._file_name

    def _feed_html_parser(self, url, parser):
        try:
            data = urlopen(url).read()
            parser.feed(astring.to_text(data, self.HTML_ENCODING))
        except HTTPError:
            raise ImageProviderError('Cannot open %s' % self.url_versions)

    def get_best_version(self, versions):
        return max(versions)

    def get_versions(self):
        """Return all available versions for the current parameters."""
        parser = VMImageHtmlParser(self.version_pattern)
        self._feed_html_parser(self.url_versions, parser)

        resulting_versions = []
        if parser.items:
            for version in parser.items:
                # Trying to convert version to int or float so max()
                # can compare numerical values.
                try:
                    # Can it be converted to integer?
                    resulting_versions.append(int(version))
                except ValueError:
                    try:
                        # Can it be converted to float?
                        resulting_versions.append(float(version))
                    except ValueError:
                        # So it's just a string
                        resulting_versions.append(version)

        return resulting_versions

    def get_version(self):
        """Probes the higher version available for the current parameters."""
        resulting_versions = self.get_versions()
        if resulting_versions:
            self._best_version = self.get_best_version(resulting_versions)
            return self._best_version
        else:
            raise ImageProviderError('Version not available at %s' %
                                     self.url_versions)

    def get_image_url(self):
        """
        Probes the higher image available for the current parameters.
        """
        if not self.url_images or not self.image_pattern:
            raise ImageProviderError(
                "url_images and image_pattern attributes are required to get image url")

        url_images = self.url_images.format(version=self.version,
                                            build=self.build,
                                            arch=self.arch)
        image = self.image_pattern.format(version=self.version,
                                          build=self.build,
                                          arch=self.arch)
        parser = VMImageHtmlParser(image)

        self._feed_html_parser(url_images, parser)

        if parser.items:
            return url_images + max(parser.items)
        else:
            raise ImageProviderError("No images matching '%s' at '%s'. "
                                     "Wrong arch?" % (image, url_images))

    def get_image_parameters(self, image_file_name):
        """
        Computation of image parameters from image_pattern

        :param image_file_name: pattern with parameters
        :type image_file_name: str
        :return: dict with parameters
        :rtype: dict or None
        """
        keywords = re.split(r'\{(.*?)\}', self.image_pattern)[1::2]
        matches = re.match(self.file_name, image_file_name)

        if not matches:
            return None

        return {x: matches.group(x) for x in keywords}


class FedoraImageProviderBase(ImageProviderBase):
    """
    Base Fedora Image Provider
    """

    HTML_ENCODING = 'iso-8859-1'
    url_old_images = None

    def get_image_url(self):
        if int(self.version) >= 28:
            cloud = 'Cloud'
        else:
            cloud = 'CloudImages'

        if self.url_old_images and int(self.version) <= 31:
            self.url_versions = self.url_old_images

        self.url_images = self.url_versions + '{version}/' \
            + cloud + '/{arch}/images/'
        return super(FedoraImageProviderBase, self).get_image_url()


class FedoraImageProvider(FedoraImageProviderBase):
    """
    Fedora Image Provider
    """

    name = 'Fedora'

    def __init__(self, version='[0-9]+', build='[0-9]+.[0-9]+',
                 arch=DEFAULT_ARCH):
        super(FedoraImageProvider, self).__init__(version, build, arch)
        self.url_versions = 'https://dl.fedoraproject.org/pub/fedora/linux/releases/'
        self.url_old_images = 'https://archives.fedoraproject.org/pub/archive/fedora/linux/releases/'
        self.image_pattern = 'Fedora-Cloud-Base-(?P<version>{version})-(?P<build>{build}).(?P<arch>{arch}).qcow2$'


class FedoraSecondaryImageProvider(FedoraImageProviderBase):
    """
    Fedora Secondary Image Provider
    """

    name = 'FedoraSecondary'

    def __init__(self, version='[0-9]+', build='[0-9]+.[0-9]+',
                 arch=DEFAULT_ARCH):
        super(FedoraSecondaryImageProvider, self).__init__(version, build,
                                                           arch)
        self.url_versions = 'https://dl.fedoraproject.org/pub/fedora-secondary/releases/'
        self.url_old_images = 'https://archives.fedoraproject.org/pub/archive/fedora-secondary/releases/'
        self.image_pattern = 'Fedora-Cloud-Base-(?P<version>{version})-(?P<build>{build}).(?P<arch>{arch}).qcow2$'


class CentOSImageProvider(ImageProviderBase):
    """
    CentOS Image Provider
    """

    name = 'CentOS'

    def __init__(self, version='[0-9]+', build='[0-9]{4}', arch=DEFAULT_ARCH):
        super(CentOSImageProvider, self).__init__(version, build, arch)
        self.url_versions = 'https://cloud.centos.org/centos/'
        self.url_images = self.url_versions + '{version}/images/'
        self.image_pattern = 'CentOS-(?P<version>{version})-(?P<arch>{arch})-GenericCloud-(?P<build>{build}).qcow2.xz$'

    def get_image_url(self):
        if int(self.version) >= 8:
            self.build = r'[\d\.\-]+'
            self.url_images = self.url_versions + '{version}/{arch}/images/'
            self.image_pattern = 'CentOS-(?P<version>{version})-GenericCloud-(?P<build>{build}).(?P<arch>{arch}).qcow2$'
        return super(CentOSImageProvider, self).get_image_url()


class UbuntuImageProvider(ImageProviderBase):
    """
    Ubuntu Image Provider
    """

    name = 'Ubuntu'

    def __init__(self, version='[0-9]+.[0-9]+', build=None,
                 arch=DEFAULT_ARCH):
        # Ubuntu uses 'amd64' instead of 'x86_64'
        if arch == 'x86_64':
            arch = 'amd64'
        # and 'arm64' instead of 'aarch64'
        elif arch == 'aarch64':
            arch = 'arm64'

        super(UbuntuImageProvider, self).__init__(version, build, arch)
        self.url_versions = 'http://cloud-images.ubuntu.com/releases/'
        self.url_images = self.url_versions + 'releases/{version}/release/'
        self.image_pattern = 'ubuntu-(?P<version>{version})-server-cloudimg-(?P<arch>{arch}).img'

    def get_versions(self):
        """Return all available versions for the current parameters."""
        parser = VMImageHtmlParser(self.version_pattern)
        self._feed_html_parser(self.url_versions, parser)
        return parser.items


class DebianImageProvider(ImageProviderBase):
    """
    Debian Image Provider
    """

    name = 'Debian'

    def __init__(self, version='[0-9]+.[0-9]+.[0-9]+.*', build=None,
                 arch=DEFAULT_ARCH):
        # Debian uses 'amd64' instead of 'x86_64'
        if arch == 'x86_64':
            arch = 'amd64'
        # and 'arm64' instead of 'aarch64'
        elif arch == 'aarch64':
            arch = 'arm64'

        super(DebianImageProvider, self).__init__(version, build, arch)
        self.url_versions = 'https://cdimage.debian.org/cdimage/openstack/'
        self.url_images = self.url_versions + '{version}/'
        self.image_pattern = 'debian-(?P<version>{version})-openstack-(?P<arch>{arch}).qcow2$'


class JeosImageProvider(ImageProviderBase):
    """
    JeOS Image Provider
    """

    name = 'JeOS'

    def __init__(self, version='[0-9]+', build=None,
                 arch=DEFAULT_ARCH):
        # JeOS uses '64' instead of 'x86_64'
        if arch == 'x86_64':
            arch = '64'

        super(JeosImageProvider, self).__init__(version, build, arch)
        self.url_versions = 'https://avocado-project.org/data/assets/jeos/'
        self.url_images = self.url_versions + '{version}/'
        self.image_pattern = 'jeos-(?P<version>{version})-(?P<arch>{arch}).qcow2.xz$'


class OpenSUSEImageProvider(ImageProviderBase):
    """
    OpenSUSE Image Provider
    """

    HTML_ENCODING = 'iso-8859-1'
    name = 'OpenSUSE'

    def __init__(self, version='[0-9]{2}.[0-9]{1}', build=None, arch=DEFAULT_ARCH):
        super(OpenSUSEImageProvider, self).__init__(version, build, arch)
        self.url_versions = 'https://download.opensuse.org/repositories/Cloud:/Images:/'
        self.url_images = self.url_versions + 'Leap_{version}/images/'

        if not build:
            self.image_pattern = 'openSUSE-Leap-(?P<version>{version})-OpenStack.(?P<arch>{arch})-((.)*).qcow2$'

        else:
            self.image_pattern = 'openSUSE-Leap-(?P<version>{version})-OpenStack.' \
                                 '(?P<arch>{arch})-(?P<build>{build}).qcow2$'

    @property
    def version_pattern(self):
        return '^Leap_%s' % self._version

    @staticmethod
    def _convert_version_numbers(versions):
        """
        Return the pure version numbers

        The version pattern return Leap_15.0, Leap_42.0, Leap_XY.Z,
        but the actual version will numeric versions only
        """
        pattern = r'(^Leap_)?([0-9{2}.[0-9]{1})'
        replace = r'\2'
        return [float(re.sub(pattern, replace, str(v))) for v in versions]

    def get_versions(self):
        versions = super(OpenSUSEImageProvider, self).get_versions()
        return self._convert_version_numbers(versions)

    def get_best_version(self, versions):
        version_numbers = self._convert_version_numbers(versions)
        if str(self._version).startswith('4'):
            version_numbers = [v for v in version_numbers if v >= 40.0]
        else:
            version_numbers = [v for v in version_numbers if v < 40.0]
        return max(version_numbers)


class CirrOSImageProvider(ImageProviderBase):
    """
    CirrOS Image Provider

    CirrOS is a Tiny OS that specializes in running on a cloud.
    """

    name = 'CirrOS'

    def __init__(self, version=r'[0-9]+\.[0-9]+\.[0-9]+', build=None, arch=DEFAULT_ARCH):
        super(CirrOSImageProvider, self).__init__(version=version, build=build, arch=arch)
        self.url_versions = 'https://download.cirros-cloud.net/'
        self.url_images = self.url_versions + '{version}/'
        self.image_pattern = 'cirros-{version}-{arch}-disk.img$'


class Image:
    def __init__(self, name, url, version, arch, build, checksum, algorithm,
                 cache_dir, snapshot_dir=None):
        """
        Creates an instance of Image class.

        :param name: Name of image.
        :type name: str
        :param url: The url where the image can be fetched from.
        :type url: str
        :param version: Version of image.
        :type version: int
        :param arch: Architecture of the system image.
        :type arch: str
        :param build: Build of the system image.
        :type build: str
        :param checksum: Hash of the system image to match after download.
        :type checksum: str
        :param algorithm: Hash type, used when the checksum is provided.
        :type algorithm: str
        :param cache_dir: Local system path where the base images will be held.
        :type cache_dir: str or iterable
        :param snapshot_dir: Local system path where the snapshot images
                            will be held.  Defaults to cache_dir if none is given.
        :type snapshot_dir: str
        """
        self.name = name
        self.url = url
        self.version = version
        self.arch = arch
        self.build = build
        self.checksum = checksum
        self.algorithm = algorithm
        self.cache_dir = cache_dir
        self.snapshot_dir = snapshot_dir

        self._path = None
        self._base_image = None

    @property
    def path(self):
        return self._path or self.get()

    @property
    def base_image(self):
        return self._base_image or self.download()

    def __repr__(self):
        return "<%s name=%s version=%s arch=%s>" % (self.__class__.__name__,
                                                    self.name,
                                                    self.version,
                                                    self.arch)

    def download(self):
        metadata = {"type": "vmimage", "name": self.name,
                    "version": self.version, "arch": self.arch,
                    "build": self.build}
        if isinstance(self.cache_dir, str):
            cache_dirs = [self.cache_dir]
        else:
            cache_dirs = self.cache_dir
        asset_path = asset.Asset(name=self.url,
                                 asset_hash=self.checksum,
                                 algorithm=self.algorithm,
                                 locations=None,
                                 cache_dirs=cache_dirs,
                                 expire=None,
                                 metadata=metadata).fetch()

        if archive.is_archive(asset_path):
            uncompressed_path = os.path.splitext(asset_path)[0]
            asset_path = archive.uncompress(asset_path, uncompressed_path)
        self._base_image = asset_path
        return self._base_image

    def get(self):
        self.download()
        self._path = self._take_snapshot()
        return self._path

    def _take_snapshot(self):
        """
        Takes a snapshot from the current image.
        """
        if QEMU_IMG is None:
            qemu_img = utils_path.find_command('qemu-img')
        else:
            qemu_img = QEMU_IMG
        name, extension = os.path.splitext(self.base_image)
        new_image = '%s-%s%s' % (name, str(uuid.uuid4()).split('-')[0],
                                 extension)
        if self.snapshot_dir is not None:
            new_image = os.path.join(self.snapshot_dir,
                                     os.path.basename(new_image))
        cmd = '%s create -f qcow2 -b %s %s' % (qemu_img,
                                               self.base_image,
                                               new_image)
        process.run(cmd)
        return new_image


def get(name=None, version=None, build=None, arch=None, checksum=None,
        algorithm=None, cache_dir=None, snapshot_dir=None):
    """
    Wrapper to get the best Image Provider, according to the parameters
    provided.

    :param name: (optional) Name of the Image Provider, usually matches
                 the distro name.
    :param version: (optional) Version of the system image.
    :param build: (optional) Build number of the system image.
    :param arch: (optional) Architecture of the system image.
    :param checksum: (optional) Hash of the system image to match after
                     download.
    :param algorithm: (optional) Hash type, used when the checksum is
                      provided.
    :param cache_dir: (optional) Local system path where the base
                      images will be held.
    :param snapshot_dir: (optional) Local system path where the snapshot images
                         will be held.  Defaults to cache_dir if none is given.

    :returns: Image instance that can provide the image
              according to the parameters.
    """
    provider = get_best_provider(name, version, build, arch)

    if cache_dir is None:
        cache_dir = tempfile.gettempdir()
    try:
        return Image(name=provider.name, url=provider.get_image_url(),
                     version=provider.version, arch=provider.arch,
                     checksum=checksum, algorithm=algorithm,
                     build=provider.build, cache_dir=cache_dir,
                     snapshot_dir=snapshot_dir)
    except ImageProviderError:
        pass

    raise AttributeError('Provider not available')


def get_best_provider(name=None, version=None, build=None, arch=None):
    """
    Wrapper to get parameters of the best Image Provider, according to
    the parameters provided.

    :param name: (optional) Name of the Image Provider, usually matches
                 the distro name.
    :param version: (optional) Version of the system image.
    :param build: (optional) Build number of the system image.
    :param arch: (optional) Architecture of the system image.

    :returns: Image Provider
    """

    if name is not None:
        name = name.lower()

    provider_args = {}
    if version is not None:
        provider_args['version'] = version
    if build is not None:
        provider_args['build'] = build
    if arch is not None:
        provider_args['arch'] = arch
        if name == 'fedora' and arch in ('ppc64', 'ppc64le', 's390x'):
            name = 'fedorasecondary'

    for provider in IMAGE_PROVIDERS:
        if name is None or name == provider.name.lower():
            try:
                return provider(**provider_args)
            except ImageProviderError:
                pass

    raise AttributeError('Provider not available')


def list_providers():
    """
    List the available Image Providers
    """
    return set(_ for _ in globals().values()
               if (isinstance(_, type) and
                   issubclass(_, ImageProviderBase) and
                   hasattr(_, 'name')))


#: List of available providers classes
IMAGE_PROVIDERS = list_providers()



import glob
import os
import tempfile
import unittest

from avocado.core import exit_codes
from avocado.utils import process, script

from .. import AVOCADO, BASEDIR, TestCaseTmpDir


class DiffTests(TestCaseTmpDir):

    def setUp(self):
        super(DiffTests, self).setUp()
        test = script.make_script(os.path.join(self.tmpdir.name, 'test'), 'exit 0')
        cmd_line = ('%s run %s '
                    '--external-runner /bin/bash '
                    '--job-results-dir %s --disable-sysinfo --json -' %
                    (AVOCADO, test, self.tmpdir.name))
        expected_rc = exit_codes.AVOCADO_ALL_OK
        self.run_and_check(cmd_line, expected_rc)
        self.jobdir = ''.join(glob.glob(os.path.join(self.tmpdir.name, 'job-*')))

        self.tmpdir2 = tempfile.TemporaryDirectory(prefix=self.tmpdir.name)
        cmd_line = ('%s run %s '
                    '--external-runner /bin/bash '
                    '--job-results-dir %s --disable-sysinfo --json -' %
                    (AVOCADO, test, self.tmpdir2.name))
        expected_rc = exit_codes.AVOCADO_ALL_OK
        self.run_and_check(cmd_line, expected_rc)
        self.jobdir2 = ''.join(glob.glob(os.path.join(self.tmpdir2.name, 'job-*')))

    def run_and_check(self, cmd_line, expected_rc):
        result = process.run(cmd_line, ignore_status=True)
        self.assertEqual(result.exit_status, expected_rc,
                         "Command %s did not return rc "
                         "%d:\n%s" % (cmd_line, expected_rc, result))
        return result

    @unittest.skipIf(os.environ.get("RUNNING_COVERAGE"), "Running coverage")
    def test_diff(self):
        cmd_line = ('%s diff %s %s' %
                    (AVOCADO, self.jobdir, self.jobdir2))
        expected_rc = exit_codes.AVOCADO_ALL_OK
        result = self.run_and_check(cmd_line, expected_rc)
        # Avocado will see the main module on the command line
        avocado_in_log = os.path.join(BASEDIR, 'avocado', '__main__.py')
        self.assertIn(b"# COMMAND LINE", result.stdout)
        self.assertIn("-%s run" % avocado_in_log, result.stdout_text)
        self.assertIn("+%s run" % avocado_in_log, result.stdout_text)

    def test_diff_nocmdline(self):
        cmd_line = ('%s diff %s %s --diff-filter nocmdline' %
                    (AVOCADO, self.jobdir, self.jobdir2))
        expected_rc = exit_codes.AVOCADO_ALL_OK
        result = self.run_and_check(cmd_line, expected_rc)
        self.assertNotIn(b"# COMMAND LINE", result.stdout)

    def tearDown(self):
        self.tmpdir.cleanup()
        self.tmpdir2.cleanup()


if __name__ == '__main__':
    unittest.main()



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat Inc. 2014
# Copyright: Jaime Huerta-Cepas <jhcepas@gmail.com> 2009
#
# Authors: Ruda Moura <rmoura@redhat.com>
#          Lucas Meneghel Rodrigues <lmr@redhat.com>
#          Jaime Huerta-Cepas <jhcepas@gmail.com>
#

"""
Tree data structure with nodes.

This tree structure (Tree drawing code) was inspired in the base tree data
structure of the ETE 2 project:

http://pythonhosted.org/ete2/

A library for analysis of phylogenetics trees.

Explicit permission has been given by the copyright owner of ETE 2
Jaime Huerta-Cepas <jhcepas@gmail.com> to take ideas/use snippets from his
original base tree code and re-license under GPLv2+, given that GPLv3 and GPLv2
(used in some avocado files) are incompatible.
"""

import collections
import copy
import itertools
import locale

from ..utils import astring


class FilterSet(set):

    """ Set of filters in standardized form """

    @staticmethod
    def __normalize(item):
        if not item.endswith("/"):
            item = item + "/"
        return item

    def add(self, item):
        return super(FilterSet, self).add(self.__normalize(item))

    def update(self, items):
        return super(FilterSet, self).update([self.__normalize(item)
                                              for item in items])

    def __str__(self):
        return ('FilterSet([%s])'
                % ', '.join(sorted(["'%s'" % i for i in self])))


class TreeEnvironment(dict):

    """ TreeNode environment with values, origins and filters """

    def __init__(self):
        super(TreeEnvironment, self).__init__()     # values
        self.origin = {}    # origins of the values
        self.filter_only = FilterSet()   # list of filter_only
        self.filter_out = FilterSet()    # list of filter_out

    def copy(self):
        cpy = TreeEnvironment()
        cpy.update(self)
        cpy.origin = copy.copy(self.origin)
        cpy.filter_only = copy.copy(self.filter_only)
        cpy.filter_out = copy.copy(self.filter_out)
        return cpy

    def __str__(self):
        """
        String representation using __str__ on items to improve readability
        """
        return self.to_text(False)

    def to_text(self, sort=False):
        """
        Human readable representation

        :param sort: Sorted to provide stable output
        :rtype: str
        """
        if sort:
            sort_fn = sorted
        else:
            def sort_fn(x):
                return x

        # Use __str__ instead of __repr__ to improve readability
        if self:
            _values = ["%s: %s" % _ for _ in sort_fn(list(self.items()))]
            values = "{%s}" % ", ".join(_values)
            _origin = ["%s: %s" % (key, node.path)
                       for key, node in sort_fn(list(self.origin.items()))]
            origin = "{%s}" % ", ".join(_origin)
        else:
            values = "{}"
            origin = "{}"
        return ",".join((values, origin, astring.to_text(self.filter_only),
                         astring.to_text(self.filter_out)))


class TreeNodeEnvOnly:

    """
    Minimal TreeNode-like class providing interface for AvocadoParams
    """

    def __init__(self, path, environment=None):
        """
        :param path: Path of this node (must not end with '/')
        :param environment: List of pair/key/value items
        """
        self.name = path.rsplit("/")[-1]
        self.path = path
        self.environment = TreeEnvironment()
        if environment:
            self.__load_environment(environment)

    def __load_environment(self, environment):
        nodes = {}
        for path, key, value in environment:
            self.environment[key] = value
            if path not in nodes:
                nodes[path] = TreeNodeEnvOnly(path)
            self.environment.origin[key] = nodes[path]

    def __eq__(self, other):
        if self.name != other.name:
            return False
        if self.path != other.path:
            return False
        if self.environment != other.environment:
            return False
        return True

    def fingerprint(self):
        return "%s%s" % (self.path, self.environment.to_text(True))

    def get_environment(self):
        return self.environment

    def get_path(self):
        return self.path


class TreeNode:

    """
    Class for bounding nodes into tree-structure.
    """

    def __init__(self, name='', value=None, parent=None, children=None):
        """
        :param name: a name for this node that will be used to define its
                     path according to the name of its parents
        :type name: str
        :param value: a collection of keys and values that will be made into
                      this node environment.
        :type value: dict
        :param parent: the node that is directly above this one in the tree
                       structure
        :type parent: :class:`TreeNode`
        :param children: the nodes that are directly beneath this one in the
                         tree structure
        :type children: builtin.list
        """
        if value is None:
            value = {}
        if children is None:
            children = []
        self.name = name
        self.value = value
        self.filters = [], []  # This node's filters, full filters are in env
        self.parent = parent
        self.children = []
        self._environment = None
        for child in children:
            self.add_child(child)

    def __repr__(self):
        return 'TreeNode(name=%r)' % self.name

    def __str__(self):
        variables = ['%s=%s' % (k, v) for k, v in self.environment.items()]
        return '%s: %s' % (self.path, ', '.join(variables))

    def __len__(self):
        """ Return number of descended leaf nodes """
        return len(tuple(self.iter_leaves()))

    def __iter__(self):
        """ Iterate through descended leaf nodes """
        return self.iter_leaves()

    def __eq__(self, other):
        """ Compares node to other node or string to name of this node """
        if isinstance(other, str):  # Compare names
            if self.name == other:
                return True
        else:
            for attr in ('name', 'value', 'children'):
                if getattr(self, attr) != getattr(other, attr):
                    return False
            return True

    def __ne__(self, other):
        """ Inverted eq """
        return not self == other

    def __hash__(self):
        values = []
        for item in self.value:
            try:
                values.append(hash(item))
            except TypeError:
                values.append(hash(astring.to_text(item)))
        children = []
        for item in self.children:
            try:
                children.append(hash(item))
            except TypeError:
                children.append(hash(astring.to_text(item)))
        return hash((self.name, ) + tuple(values) + tuple(children))

    def fingerprint(self):
        """
        Reports string which represents the value of this node.
        """
        return "%s%s" % (self.path, self.environment.to_text(True))

    def add_child(self, node):
        """
        Append node as child. Nodes with the same name gets merged into the
        existing position.
        """
        if isinstance(node, TreeNode):
            if node.name in self.children:
                self.children[self.children.index(node.name)].merge(node)
            else:
                node.parent = self
                self.children.append(node)
        else:
            raise ValueError('Bad node type.')

    def merge(self, other):
        """
        Merges `other` node into this one without checking the name of the
        other node. New values are appended, existing values overwritten
        and unaffected ones are kept. Then all other node children are
        added as children (recursively they get either appended at the end
        or merged into existing node in the previous position.
        """
        self.value.update(other.value)
        self.filters[0].extend(other.filters[0])
        self.filters[1].extend(other.filters[1])
        for child in other.children:
            self.add_child(child)

    @property
    def is_leaf(self):
        """ Is this a leaf node? """
        return not self.children

    @property
    def root(self):
        """ Root of this tree """
        return self.get_root()

    def get_root(self):
        """ Get root of this tree """
        root = self
        for root in self.iter_parents():
            pass
        return root

    def iter_parents(self):
        """ Iterate through parent nodes to root """
        node = self.parent
        while True:
            if node is None:
                return
            yield node
            node = node.parent

    @property
    def parents(self):
        """ List of parent nodes """
        return self.get_parents()

    def get_parents(self):
        """ Get list of parent nodes """
        return list(self.iter_parents())

    @property
    def path(self):
        """ Node path """
        return self.get_path()

    def get_path(self, sep='/'):
        """ Get node path """
        if not self.parent:
            return sep + astring.to_text(self.name)
        path = [astring.to_text(self.name)]
        for node in self.iter_parents():
            path.append(astring.to_text(node.name))
        return sep.join(reversed(path))

    @property
    def environment(self):
        """ Node environment (values + preceding envs) """
        return self.get_environment()

    def get_environment(self):
        """ Get node environment (values + preceding envs) """
        if self._environment is None:
            self._environment = (self.parent.environment.copy()
                                 if self.parent else TreeEnvironment())
            for key, value in self.value.items():
                if isinstance(value, list):
                    if (key in self._environment and
                            isinstance(self._environment[key], list)):
                        self._environment[key] = self._environment[key] + value
                    else:
                        self._environment[key] = value
                else:
                    self._environment[key] = value
                self._environment.origin[key] = self
            self._environment.filter_only.update(self.filters[0])
            self._environment.filter_out.update(self.filters[1])
        return self._environment

    def set_environment_dirty(self):
        """
        Set the environment cache dirty. You should call this always when
        you query for the environment and then change the value or structure.
        Otherwise you'll get the old environment instead.
        """
        for child in self.children:
            child.set_environment_dirty()
        self._environment = None

    def get_node(self, path, create=False):
        """
        :param path: Path of the desired node (relative to this node)
        :param create: Create the node (and intermediary ones) when not present
        :return: the node associated with this path
        :raise ValueError: When path doesn't exist and create not set
        """
        node = self
        for name in path.split('/'):
            if not name:
                continue
            try:
                node = node.children[node.children.index(name)]
            except ValueError:
                if create:
                    child = node.__class__(name)
                    node.add_child(child)
                    node = child
                else:
                    raise ValueError("Path %s does not exists in this tree\n%s"
                                     % (path, tree_view(self.root)))
        return node

    def iter_children_preorder(self):
        """ Iterate through children """
        queue = collections.deque()
        node = self
        while node is not None:
            yield node
            queue.extendleft(reversed(node.children))
            try:
                node = queue.popleft()
            except IndexError:
                node = None

    def iter_leaves(self):
        """ Iterate through leaf nodes """
        for node in self.iter_children_preorder():
            if node.is_leaf:
                yield node

    def get_leaves(self):
        """ Get list of leaf nodes """
        return list(self.iter_leaves())

    def detach(self):
        """ Detach this node from parent """
        if self.parent:
            self.parent.children.remove(self)
            self.parent = None
        return self


def tree_view(root, verbose=None, use_utf8=None):
    """
    Generate tree-view of the given node
    :param root: root node
    :param verbose: verbosity (0, 1, 2, 3)
    :param use_utf8: Use utf-8 encoding (None=autodetect)
    :return: string representing this node's tree structure
    """

    def prefixed_write(prefix1, prefix2, value):
        """
        Split value's lines and prepend empty prefix to 2nd+ lines
        :return: list of lines
        """
        value = astring.to_text(value)
        if '\n' not in value:
            return [prefix1 + prefix2 + value]
        value = value.splitlines()
        empty_prefix2 = ' ' * len(prefix2)
        return [prefix1 + prefix2 + value[0]] + [prefix1 + empty_prefix2 +
                                                 _ for _ in value[1:]]

    def process_node(node):
        """
        Generate this node's tree-view
        :return: list of lines
        """
        if getattr(node, "multiplex", None):
            down = charset['DoubleDown']
            down_right = charset['DoubleDownRight']
            right = charset['DoubleRight']
        else:
            down = charset['Down']
            down_right = charset['DownRight']
            right = charset['Right']
        out = [node.name]
        if verbose is not None and verbose >= 2 and node.is_leaf:
            values = itertools.chain(iter(node.environment.items()),
                                     [("filter-only", _)
                                      for _ in node.environment.filter_only],
                                     [("filter-out", _)
                                      for _ in node.environment.filter_out])
        elif verbose in (1, 3):
            values = itertools.chain(iter(node.value.items()),
                                     [("filter-only", _)
                                      for _ in node.filters[0]],
                                     [("filter-out", _)
                                      for _ in node.filters[1]])
        else:
            values = None
        if values:
            val = charset['Value']
            if node.children:
                val_prefix = down
            else:
                val_prefix = '  '
            for key, value in values:
                out.extend(prefixed_write(val_prefix, "%s%s: " % (val, key),
                                          value))
        if node.children:
            for child in node.children[:-1]:
                lines = process_node(child)
                out.append(down_right + lines[0])
                out.extend(down + line for line in lines[1:])
            lines = process_node(node.children[-1])
            out.append(right + lines[0])
            empty_down_right = ' ' * len(down_right)
            out.extend(empty_down_right + line for line in lines[1:])
        return out

    if use_utf8 is None:
        use_utf8 = locale.getdefaultlocale()[1] == 'UTF-8'
    if use_utf8:
        charset = {'DoubleDown': u' \u2551   ',
                   'DoubleDownRight': u' \u2560\u2550\u2550 ',
                   'DoubleRight': u' \u255a\u2550\u2550 ',
                   'Down': u' \u2503   ',
                   'DownRight': u' \u2523\u2501\u2501 ',
                   'Right': u' \u2517\u2501\u2501 ',
                   'Value': u'\u2192 '}
    else:   # ASCII fallback
        charset = {'Down': ' |   ',
                   'DownRight': ' |-- ',
                   'Right': ' \\-- ',
                   'DoubleDown': ' #   ',
                   'DoubleDownRight': ' #== ',
                   'DoubleRight': ' #== ',
                   'Value': ' -> '}
    if getattr(root, "multiplex", None):
        down = charset['DoubleDown']
        down_right = charset['DoubleDownRight']
        right = charset['DoubleRight']
    else:
        down = charset['Down']
        down_right = charset['DownRight']
        right = charset['Right']
    out = []
    if verbose is not None and verbose >= 2 and root.is_leaf:
        values = root.environment.items()
    elif verbose in (1, 3):
        values = root.value.items()
    else:
        values = None
    if values:
        prefix = charset['Value'].lstrip()
        for key, value in values:
            out.extend(prefixed_write(prefix, key + ': ', value))
    if root.children:
        for child in root.children[:-1]:
            lines = process_node(child)
            out.append(down_right + lines[0])
            out.extend(down + line for line in lines[1:])
        lines = process_node(root.children[-1])
        out.append(right + lines[0])
        out.extend(' ' * len(down_right) + line for line in lines[1:])
    # When not on TTY we need to force the encoding
    return '\n'.join(out).encode('utf-8' if use_utf8 else 'ascii',
                                 errors='xmlcharrefreplace')



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat Inc. 2013-2014
# Authors: Lucas Meneghel Rodrigues <lmr@redhat.com>
#          Ruda Moura <rmoura@redhat.com>

"""
Contains the Result class, used for result accounting.
"""


class Result:

    """
    Result class, holder for job (and its tests) result information.
    """

    def __init__(self, job_unique_id, job_logfile):
        """
        Creates an instance of Result.

        :param job_unique_id: the job's unique ID, usually from
                              :attr:`avocado.core.job.Job.unique_id`
        :param job_logfile: the job's unique ID, usually from
                            :attr:`avocado.core.job.Job.logfile`
        """
        self.job_unique_id = job_unique_id
        self.logfile = job_logfile
        self.tests_total = 0
        self.tests_run = 0
        self.tests_total_time = 0.0
        self.passed = 0
        self.errors = 0
        self.failed = 0
        self.skipped = 0
        self.warned = 0
        self.interrupted = 0
        self.cancelled = 0
        self.tests = []

    def _reconcile(self):
        """
        Make sure job results are reconciled

        In situations such as job interruptions, some test results will be
        missing, but this is no excuse for giving wrong summaries of test
        results.
        """
        valid_results_count = (self.passed + self.errors +
                               self.failed + self.warned +
                               self.skipped + self.interrupted +
                               self.cancelled)
        other_skipped_count = self.tests_total - valid_results_count
        if other_skipped_count > 0:
            self.skipped += other_skipped_count
        else:
            self.tests_total -= other_skipped_count

    def end_tests(self):
        """
        Called once after all tests are executed.
        """
        self._reconcile()

    def start_test(self, state):
        """
        Called when the given test is about to run.

        :param state: result of :class:`avocado.core.test.Test.get_state`.
        :type state: dict
        """

    def end_test(self, state):
        """
        Called when the given test has been run.

        :param state: result of :class:`avocado.core.test.Test.get_state`.
        :type state: dict
        """
        self.tests_run += 1
        self.tests_total_time += state.get('time_elapsed', -1)
        self.tests.append(state)

    def check_test(self, state):
        """
        Called once for a test to check status and report.

        :param test: A dict with test internal state
        """
        status = state.get('status')
        if status == "PASS":
            self.passed += 1
        elif status == "SKIP":
            self.skipped += 1
        elif status == "FAIL":
            self.failed += 1
        elif status == "WARN":
            self.warned += 1
        elif status == "INTERRUPTED":
            self.interrupted += 1
        elif status == "CANCEL":
            self.cancelled += 1
        else:
            self.errors += 1
        self.end_test(state)

    @property
    def rate(self):
        total = float(self.tests_total - self.skipped - self.cancelled)
        if not total:
            return 0.0
        succeeded = float(self.passed + self.warned)
        return 100 * (succeeded / total)
